# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel, info@privacyidea.org
#  http://www.privacyidea.org
#
#     2014-06-30    Cornelius Kölbel, added tables clientmachine,
#                                                clienttoken,
#                                                clientoptions
#
#     2014-08-01    Cornelius Kölbel, added table machinetoken, machineuser,
#                                                 and corresponding options
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  LSE
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
  Description:  This is part of the privacyidea service
                This file contains the database definition / database model

  Dependencies: -


'''


import binascii
import logging
import sys
import traceback

from datetime import datetime
from datetime import date

from json import loads, dumps


"""The application's model objects"""
import sqlalchemy as sa

from sqlalchemy import orm
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relation
from sqlalchemy import UniqueConstraint

from privacyidea.model import meta
from privacyidea.model.meta import Session
from privacyidea.lib.crypto import geturandom
from privacyidea.lib.crypto import encrypt, hash, SecretObj
from privacyidea.lib.crypto import encryptPin
from privacyidea.lib.crypto import decryptPin
from privacyidea.lib.crypto import get_rand_digit_str
from privacyidea.lib.log import log_with
log = logging.getLogger(__name__)

from pylons import config
implicit_returning = config.get('privacyideaSQL.implicit_returning', True)


@log_with(log)
def init_model(engine):
    """Call me before using any of the tables or classes in the model"""

    meta.Session.configure(bind=engine)
    meta.engine = engine


token_table = sa.Table('Token', meta.metadata,
                       sa.Column('privacyIDEATokenId', sa.types.Integer(),
                                 sa.Sequence('token_seq_id', optional=True),
                                 primary_key=True, nullable=False),
                       sa.Column('privacyIDEATokenDesc', sa.types.Unicode(80),
                                 default=u''),
                       sa.Column('privacyIDEATokenSerialnumber',
                                 sa.types.Unicode(40), default=u'',
                                 unique=True, nullable=False, index=True),
                       sa.Column('privacyIDEATokenType', sa.types.Unicode(30),
                                 default=u'HMAC', index=True),
                       sa.Column('privacyIDEATokenInfo',
                                 sa.types.Unicode(2000), default=u''),
                       sa.Column('privacyIDEATokenPinUser',
                                 sa.types.Unicode(512),
                                 default=u''),  # encrypt
                       sa.Column('privacyIDEATokenPinUserIV',
                                 sa.types.Unicode(32), default=u''),  # encrypt
                       sa.Column('privacyIDEATokenPinSO',
                                 sa.types.Unicode(512),
                                 default=u''),  # encrypt
                       sa.Column('privacyIDEATokenPinSOIV',
                                 sa.types.Unicode(32),
                                 default=u''),  # encrypt

                       sa.Column('privacyIDEAIdResolver',
                                 sa.types.Unicode(120), default=u'',
                                 index=True),
                       sa.Column('privacyIDEAIdResClass',
                                 sa.types.Unicode(120), default=u''),
                       sa.Column('privacyIDEAUserid',
                                 sa.types.Unicode(320),
                                 default=u'', index=True),
                       
                       sa.Column('privacyIDEASeed', sa.types.Unicode(32),
                                 default=u''),
                       sa.Column('privacyIDEAOtpLen', sa.types.Integer(),
                                 default=6),
                       sa.Column('privacyIDEAPinHash', sa.types.Unicode(512),
                                 default=u''),  # hashed
                       sa.Column('privacyIDEAKeyEnc', sa.types.Unicode(1024),
                                 default=u''),  # encrypt
                       sa.Column('privacyIDEAKeyIV', sa.types.Unicode(32),
                                 default=u''),

                       sa.Column('privacyIDEAMaxFail', sa.types.Integer(),
                                 default=10),
                       sa.Column('privacyIDEAIsactive', sa.types.Boolean(),
                                 default=True),
                       sa.Column('privacyIDEAFailCount', sa.types.Integer(),
                                 default=0),
                       sa.Column('privacyIDEACount', sa.types.Integer(),
                                 default=0),
                       sa.Column('privacyIDEACountWindow', sa.types.Integer(),
                                 default=10),
                       sa.Column('privacyIDEASyncWindow', sa.types.Integer(),
                                 default=1000),
                       implicit_returning=implicit_returning,
                       )


class Token(object):

    @log_with(log)
    def __init__(self, serial):

        # self.privacyIDEATokenId - will be generated DBType serial
        self.privacyIDEATokenSerialnumber = u'' + serial

        self.privacyIDEATokenType = u''

        self.privacyIDEACount = 0
        self.privacyIDEAFailCount = 0
        # get maxFail should have a configurable default
        self.privacyIDEAMaxFail = 10
        self.privacyIDEAIsactive = True
        self.privacyIDEACountWindow = 10
        self.privacyIDEAOtpLen = 6
        self.privacyIDEASeed = u''

        self.privacyIDEAIdResolver = None
        self.privacyIDEAIdResClass = None
        self.privacyIDEAUserid = None

        # will be assigned automaticaly
        # self.privacyIDEATokenId      = 0

    def _fix_spaces(self, data):
        '''
        On MS SQL server empty fields ("") like the privacyIDEATokenInfo
        are returned as a string with a space (" ").
        This functions helps fixing this.
        Also avoids running into errors, if the data is a None Type.

        :param data: a string from teh database
        :type data: usually a string
        :return: a stripped string
        '''
        if data:
            data = data.strip()

        return data

    def getSerial(self):
        return self.privacyIDEATokenSerialnumber

    @log_with(log)
    def setHKey(self, hOtpKey, reset_failcount=True):
        iv = geturandom(16)
        enc_otp_key = encrypt(hOtpKey, iv)
        self.privacyIDEAKeyEnc = unicode(binascii.hexlify(enc_otp_key))
        length = len(self.privacyIDEAKeyEnc)
        if length > 1024:
            log.error("Key %s exceeds database field %d!" % (self.getSerial(),
                                                             length))
        self.privacyIDEAKeyIV = unicode(binascii.hexlify(iv))
        self.privacyIDEACount = 0
        if True == reset_failcount:
            self.privacyIDEAFailCount = 0

    @log_with(log)
    def setUserPin(self, userPin):
        iv = geturandom(16)
        enc_userPin = encrypt(userPin, iv)
        self.privacyIDEATokenPinUser = unicode(binascii.hexlify(enc_userPin))
        self.privacyIDEATokenPinUserIV = unicode(binascii.hexlify(iv))

    @log_with(log)
    def getHOtpKey(self):
        key = binascii.unhexlify(self.privacyIDEAKeyEnc)
        iv = binascii.unhexlify(self.privacyIDEAKeyIV)
        secret = SecretObj(key, iv)
        return secret

    @log_with(log)
    def getOtpCounter(self):
        return self.privacyIDEACount

    @log_with(log)
    def getUserPin(self):
        pu = self.privacyIDEATokenPinUser
        if pu is None:
            pu = ''
        puiv = self.privacyIDEATokenPinUserIV
        if puiv is None:
            puiv = ''

        key = binascii.unhexlify(pu)
        iv = binascii.unhexlify(puiv)
        secret = SecretObj(key, iv)
        return secret

    def setHashedPin(self, pin):
        log.debug('setHashedPin()')
        seed = geturandom(16)
        self.privacyIDEASeed = unicode(binascii.hexlify(seed))
        self.privacyIDEAPinHash = unicode(binascii.hexlify(hash(pin, seed)))
        return self.privacyIDEAPinHash

    def getHashedPin(self, pin):
        # TODO: we could log the PIN here.
        log.debug('getHashedPin()')

        # calculate a hash from a pin
        # Fix for working with MS SQL servers
        # MS SQL servers sometimes return a '<space>' when the
        # column is empty: ''
        seed_str = self._fix_spaces(self.privacyIDEASeed)
        seed = binascii.unhexlify(seed_str)
        hPin = hash(pin, seed)
        log.debug("hPin: %s, pin: %s, seed: %s" % (binascii.hexlify(hPin),
                                                   pin,
                                                   self.privacyIDEASeed))
        return binascii.hexlify(hPin)
        
    @log_with(log)
    def setDescription(self, desc):
        if desc is None:
            desc = ""
        self.privacyIDEATokenDesc = unicode(desc)
        return self.privacyIDEATokenDesc

    @log_with(log)
    def setOtpLen(self, otplen):
        self.privacyIDEAOtpLen = int(otplen)

    def setPin(self, pin, hashed=True):
        # TODO: we could log the PIN here
        log.debug("setPin()")

        upin = ""
        if pin != "" and pin is not None:
            upin = pin
        if hashed is True:
            self.setHashedPin(upin)
            log.debug("setPin(HASH:%r)" % self.privacyIDEAPinHash)
        elif hashed is False:
            self.privacyIDEAPinHash = "@@" + encryptPin(upin)
            log.debug("setPin(ENCR:%r)" % self.privacyIDEAPinHash)
        return self.privacyIDEAPinHash

    def comparePin(self, pin):
        log.debug("entering comparePin")
        res = False

        # check for a valid input
        if pin is None:
            log.error("no valid PIN!")
            return res

        if (self.isPinEncrypted() is True):
            log.debug("we got an encrypted PIN!")
            tokenPin = self.privacyIDEAPinHash[2:]
            decryptTokenPin = decryptPin(tokenPin)
            if (decryptTokenPin == pin):
                res = True
        else:
            log.debug("we got a hashed PIN!")
            if len(self.privacyIDEAPinHash) > 0:
                mypHash = self.getHashedPin(pin)
            else:
                mypHash = pin
            if (mypHash == self.privacyIDEAPinHash):
                res = True

        return res

    @log_with(log)
    def deleteToken(self):
        # some DBs (eg. DB2) run in deadlock, if the TokenRealm entry
        # is deleteted via foreign key relation
        # so we delete it explicit
        Session.query(TokenRealm).filter(TokenRealm.token_id == self.privacyIDEATokenId).delete()
        Session.delete(self)
        return True

    def isPinEncrypted(self, pin=None):
        ret = False
        if pin is None:
            pin = self.privacyIDEAPinHash
        if (pin.startswith("@@") is True):
            ret = True
        return ret

    def getPin(self):
        ret = -1
        if self.isPinEncrypted() is True:
            tokenPin = self.privacyIDEAPinHash[2:]
            ret = decryptPin(tokenPin)
        return ret

    def setSoPin(self, soPin):
        # TODO: we could log the PIN here
        log.debug('setSoPin()')
        iv = geturandom(16)
        enc_soPin = encrypt(soPin, iv)
        self.privacyIDEATokenPinSO = unicode(binascii.hexlify(enc_soPin))
        self.privacyIDEATokenPinSOIV = unicode(binascii.hexlify(iv))

    def __unicode__(self):
        return self.privacyIDEATokenDesc

    @log_with(log)
    def get(self, key=None, fallback=None, save=False):
        '''
        simulate the dict behaviour to make challenge processing
        easier, as this will have to deal as well with
        'dict only challenges'

        :param key: the attribute name - in case of key is not provided, a dict
                    of all class attributes are returned
        :param fallback: if the attribute is not found,
                         the fallback is returned
        :param save: in case of all attributes and save==True, the timestamp is
                     converted to a string representation
        '''
        if key is None:
            return self.get_vars(save=save)

        if hasattr(self, key):
            kMethod = "get" + key.capitalize()
            if hasattr(self, kMethod):
                return getattr(self, kMethod)()
            else:
                return getattr(self, key)
        else:
            return fallback

    @log_with(log)
    def get_vars(self, save=False):
        log.debug('get_vars()')

        ret = {}
        ret['privacyIDEA.TokenId'] = self.privacyIDEATokenId
        ret['privacyIDEA.TokenDesc'] = self.privacyIDEATokenDesc
        ret['privacyIDEA.TokenSerialnumber'] = self.privacyIDEATokenSerialnumber

        ret['privacyIDEA.TokenType'] = self.privacyIDEATokenType
        ret['privacyIDEA.TokenInfo'] = self._fix_spaces(self.privacyIDEATokenInfo)
        # ret['privacyIDEATokenPinUser']   = self.privacyIDEATokenPinUser
        # ret['privacyIDEATokenPinSO']     = self.privacyIDEATokenPinSO

        ret['privacyIDEA.IdResolver'] = self.privacyIDEAIdResolver
        ret['privacyIDEA.IdResClass'] = self.privacyIDEAIdResClass
        ret['privacyIDEA.Userid'] = self.privacyIDEAUserid
        ret['privacyIDEA.OtpLen'] = self.privacyIDEAOtpLen
        # ret['privacyIDEA.PinHash']        = self.privacyIDEAPinHash

        ret['privacyIDEA.MaxFail'] = self.privacyIDEAMaxFail
        ret['privacyIDEA.Isactive'] = self.privacyIDEAIsactive
        ret['privacyIDEA.FailCount'] = self.privacyIDEAFailCount
        ret['privacyIDEA.Count'] = self.privacyIDEACount
        ret['privacyIDEA.CountWindow'] = self.privacyIDEACountWindow
        ret['privacyIDEA.SyncWindow'] = self.privacyIDEASyncWindow

        # list of Realm names
        ret['privacyIDEA.RealmNames'] = self.getRealmNames()

        return ret

    __str__ = __unicode__

    def __repr__(self):
        '''
        return the token state as text

        :return: token state as string representation
        :rtype:  string
        '''
        ldict = {}
        for attr in self.__dict__:
            key = "%r" % attr
            val = "%r" % getattr(self, attr)
            ldict[key] = val
        res = "<%r %r>" % (self.__class__, ldict)
        return res

    def getSyncWindow(self):
        return self.privacyIDEASyncWindow

    def setCountWindow(self, counter):
        self.privacyIDEACountWindow = counter

    def getCountWindow(self):
        return self.privacyIDEACountWindow

    def getInfo(self):
        # Fix for working with MS SQL servers
        # MS SQL servers sometimes return a '<space>' when
        # the column is empty: ''
        return self._fix_spaces(self.privacyIDEATokenInfo)

    def setInfo(self, info):
        self.privacyIDEATokenInfo = info

    def _setPin(self, pin, hashed=True):
        log.debug("_setPin(%s)" % pin)
        if pin is None or pin == "":
            log.debug("Token pin was None")
        else:
            self.setPin(pin, hashed)

    @log_with(log)
    def storeToken(self):
        Session.add(self)
        Session.flush()
        Session.commit()
        return True

    def setType(self, typ):
        self.privacyIDEATokenType = typ
        return

    def getType(self):
        return self.privacyIDEATokenType

    def updateType(self, typ):
        # in case the prevoius has been different type
        # we must reset the counters
        # But be aware, ray, this could also be upper and lower case mixing...
        if self.privacyIDEATokenType.lower() != typ.lower():
            self.privacyIDEACount = 0
            self.privacyIDEAFailCount = 0

        self.privacyIDEATokenType = typ
        return

    def updateOtpKey(self, otpKey):
        # in case of a new hOtpKey we have to do some more things
        if (otpKey is not None):
            secretObj = self.getHOtpKey()
            if secretObj.compare(otpKey) is False:
                log.debug('update token OtpKey - counter reset')
                self.setHKey(otpKey)

    def updateToken(self, tokenDesc, otpKey, pin):
        log.debug('updateToken()')
        self.setDescription(tokenDesc)
        self._setPin(pin)
        self.updateOtpKey(otpKey)

    def getRealms(self):
        return self.realms

    def getRealmNames(self):
        r_list = []
        for r in self.realms:
            r_list.append(r.name)
        return r_list

    def addRealm(self, realm):
        if realm is not None:
            self.realms.append(realm)
        else:
            log.error("adding empty realm!")

    def setRealms(self, realms):
        if realms is not None:
            self.realms = realms
        else:
            log.error("assigning empty realm!")


@log_with(log)
def createToken(serial):
    serial = u'' + serial
    token = Token(serial)
    return token


config_table = sa.Table('Config', meta.metadata,
                        sa.Column('Key', sa.types.Unicode(255),
                                  primary_key=True, nullable=False),
                        sa.Column('Value', sa.types.Unicode(2000),
                                  default=u''),
                        sa.Column('Type', sa.types.Unicode(2000),
                                  default=u''),
                        sa.Column('Description', sa.types.Unicode(2000),
                                  default=u''),
                        implicit_returning=implicit_returning,
                        )

log.debug('config table append_column')


class Config(object):
    
    @log_with(log)
    def __init__(self, Key, Value, Type=u'', Description=u''):
        if (not Key.startswith("privacyidea.") and
                not Key.startswith("encprivacyidea.")):
            Key = "privacyidea." + Key

        self.Key = unicode(Key)
        self.Value = unicode(Value)
        self.Type = unicode(Type)
        self.Description = unicode(Description)

    def __unicode__(self):
        return self.Description

    __str__ = __unicode__


# This table connect a token to several realms
tokenrealm_table = sa.Table('TokenRealm', meta.metadata,
                            sa.Column('id', sa.types.Integer(),
                                      sa.Sequence('tokenrealm_seq_id',
                                                  optional=True),
                                      primary_key=True,
                                      nullable=False),
                            sa.Column('token_id', sa.types.Integer(),
                                      ForeignKey('Token.privacyIDEATokenId')),
                            sa.Column('realm_id', sa.types.Integer(),
                                      ForeignKey('Realm.id')),
                            implicit_returning=implicit_returning,
                            )


class TokenRealm(object):

    def __init__(self, realmid):
        log.debug("setting realm_id to %i" % realmid)
        self.realm_id = realmid
        self.token_id = 0


realm_table = sa.Table('Realm', meta.metadata,
                       sa.Column('id', sa.types.Integer(),
                                 sa.Sequence('realm_seq_id', optional=True),
                                 primary_key=True, nullable=False),
                       sa.Column('name', sa.types.Unicode(255), default=u'',
                                 unique=True, nullable=False),
                       sa.Column('default', sa.types.Boolean(), default=False),
                       sa.Column('option', sa.types.Unicode(40), default=u''),
                       implicit_returning=implicit_returning,
                       )


class Realm(object):
    
    @log_with(log)
    def __init__(self, realm):
        self.name = realm
        # self.id     = 0

    @log_with(log)
    def storeRealm(self):
        Session.add(self)
        Session.commit()
        return True


''' ''' '''
ocra challenges are stored
''' ''' '''

log.debug('creating ocra table')

ocra_table = sa.Table('ocra', meta.metadata,
                      sa.Column('id', sa.types.Integer(),
                                sa.Sequence('ocra_seq_id', optional=True),
                                primary_key=True, nullable=False),
                      sa.Column('transid', sa.types.Unicode(20), unique=True,
                                nullable=False, index=True),
                      sa.Column('data', sa.types.Unicode(512),
                                default=u''),
                      sa.Column('challenge', sa.types.Unicode(256),
                                default=u''),
                      sa.Column('session', sa.types.Unicode(512),
                                default=u''),
                      sa.Column('tokenserial', sa.types.Unicode(64),
                                default=u''),
                      sa.Column('timestamp', sa.types.DateTime,
                                default=datetime.now()),
                      sa.Column('received_count', sa.types.Integer(),
                                default=0),
                      sa.Column('received_tan', sa.types.Boolean,
                                default=False),
                      sa.Column('valid_tan', sa.types.Boolean,
                                default=False),
                      implicit_returning=implicit_returning,
                      )


class OcraChallenge(object):
    '''
    '''
    @log_with(log)
    def __init__(self, transId, challenge, tokenserial, data, session=u''):
        
        self.transid = u'' + transId
        self.challenge = u'' + challenge
        self.tokenserial = u'' + tokenserial
        self.data = u'' + data
        self.timestamp = datetime.now()
        self.session = u'' + session
        self.received_count = 0
        self.received_tan = False
        self.valid_tan = False

    def setData(self, data):
        self.data = unicode(data)

    def getData(self):
        return self.data

    def getSession(self):
        return self.session

    def setSession(self, session):
        self.session = unicode(session)

    def setChallenge(self, challenge):
        self.challenge = unicode(challenge)

    def setTanStatus(self, received=False, valid=False):
        self.received_tan = received
        self.received_count += 1
        self.valid_tan = valid

    def getTanStatus(self):
        return (self.received_tan, self.valid_tan)

    def getChallenge(self):
        return self.challenge

    def getTransactionId(self):
        return self.transid

    @log_with(log)
    def save(self):
        Session.add(self)
        Session.commit()
        return self.transid

    def __unicode__(self):
        descr = {}
        descr['id'] = self.id
        descr['transid'] = self.transid
        descr['challenge'] = self.challenge
        descr['tokenserial'] = self.tokenserial
        descr['data'] = self.data
        descr['timestamp'] = self.timestamp
        descr['received_tan'] = self.received_tan
        descr['valid_tan'] = self.valid_tan

        return "%s" % unicode(descr)

    __str__ = __unicode__


''' ''' '''
challenges are stored
''' ''' '''

log.debug('creating challenges table')

challenges_table = sa.Table('challenges', meta.metadata,
                            sa.Column('id', sa.types.Integer(),
                                      sa.Sequence('challenges_seq_id',
                                                  optional=True),
                                      primary_key=True, nullable=False),
                            sa.Column('transid', sa.types.Unicode(64),
                                      unique=True, nullable=False,
                                      index=True),
                            sa.Column('data', sa.types.Unicode(512),
                                      default=u''),
                            sa.Column('challenge', sa.types.Unicode(512),
                                      default=u''),
                            sa.Column('session', sa.types.Unicode(512),
                                      default=u''),
                            sa.Column('tokenserial', sa.types.Unicode(64),
                                      default=u''),
                            sa.Column('timestamp', sa.types.DateTime,
                                      default=datetime.now()),
                            sa.Column('received_count', sa.types.Integer(),
                                      default=0),
                            sa.Column('received_tan', sa.types.Boolean,
                                      default=False),
                            sa.Column('valid_tan', sa.types.Boolean,
                                      default=False),
                            implicit_returning=implicit_returning,
                            )


class Challenge(object):
    '''
    the generic challenge handling
    '''
    @log_with(log)
    def __init__(self, transid, tokenserial,
                 challenge=u'', data=u'', session=u''):

        self.transid = u'' + transid
        self.challenge = u'' + challenge
        self.tokenserial = u'' + tokenserial
        self.data = u'' + data
        self.timestamp = datetime.now()
        self.session = u'' + session
        self.received_count = 0
        self.received_tan = False
        self.valid_tan = False

    @classmethod
    def createTransactionId(cls, length=20):
        return get_rand_digit_str(length)

    def setData(self, data):
        if type(data) in [dict, list]:
            self.data = dumps(data)
        else:
            self.data = unicode(data)

    def get(self, key=None, fallback=None, save=False):
        '''
        simulate the dict behaviour to make challenge processing
        easier, as this will have to deal as well with
        'dict only challenges'

        :param key: the attribute name - in case of key is not provided, a dict
                    of all class attributes are returned
        :param fallback: if the attribute is not found, the
                         fallback is returned
        :param save: in case of all attributes and save==True, the timestamp is
                     converted to a string representation
        '''
        if key is None:
            return self.get_vars(save=save)

        if hasattr(self, key):
            kMethod = "get" + key.capitalize()
            if hasattr(self, kMethod):
                return getattr(self, kMethod)()
            else:
                return getattr(self, key)
        else:
            return fallback

    def getId(self):
        return self.id

    def getData(self):
        data = {}
        try:
            data = loads(self.data)
        except:
            data = self.data
        return data

    def getSession(self):
        return self.session

    def setSession(self, session):
        self.session = unicode(session)

    def setChallenge(self, challenge):
        self.challenge = unicode(challenge)

    def setTanStatus(self, received=False, valid=False):
        self.received_tan = received
        self.received_count += 1
        self.valid_tan = valid

    def getTanStatus(self):
        return (self.received_tan, self.valid_tan)

    def getTanCount(self):
        return self.received_count

    def getChallenge(self):
        return self.challenge

    def getTransactionId(self):
        return self.transid

    def getTokenSerial(self):
        return self.tokenserial

    @log_with(log)
    def save(self):
        '''
        enforce the saveing of a challenge
        - will guarentee the uniqness of the transaction id

        :return: transaction id of the stored challeng
        '''
        try:
            Session.add(self)
            Session.commit()
            log.debug('save challenge : success')

        except Exception as exce:
            log.error('Error during saving challenge: %r' % exce)
            log.error("%s" % traceback.format_exc())

        return self.transid

    def get_vars(self, save=False):
        '''
        return a dictionary of all vars in the challenge class

        :return: dict of vars
        '''
        descr = {}
        descr['id'] = self.id
        descr['transid'] = self.transid
        descr['challenge'] = self.challenge
        descr['tokenserial'] = self.tokenserial
        descr['data'] = self.getData()
        if save is True:
            descr['timestamp'] = "%s" % self.timestamp
        else:
            descr['timestamp'] = self.timestamp
        descr['received_tan'] = self.received_tan
        descr['valid_tan'] = self.valid_tan
        return descr

    def __unicode__(self):
        descr = self.get_vars()
        return "%s" % unicode(descr)

    __str__ = __unicode__

#
# Client machine table
# This is the table that holds the definition of a client machine

clientmachine_table = sa.Table('ClientMachine', meta.metadata,
                               sa.Column('id', sa.types.Integer(),
                                         sa.Sequence('clientmachine_seq_id',
                                                     optional=True),
                                         primary_key=True, nullable=False),
                               sa.Column("cm_ip", sa.types.Unicode(64),
                                         sa.Sequence('clientmachine_seq_ip',
                                                     optional=True),
                                         unique=True, nullable=False),
                               sa.Column("cm_name", sa.types.Unicode(64),
                                         sa.Sequence('clientmachine_seq_name',
                                                     optional=True),
                                         unique=True, nullable=False),
                               sa.Column("cm_desc", sa.types.Unicode(128)),
                               sa.Column("cm_decommission", sa.types.DateTime),
                               UniqueConstraint('id', 'cm_ip',
                                                'cm_name', name='uix_2'),
                               implicit_returning=implicit_returning,
                               )


class Machine(object):
    
    @log_with(log)
    def __init__(self, name, ip=u'',
                 desc=u'',
                 decommission=None):
        '''
        Create a new Machine database object
        
        :param decommission: This is the date, when the machine should be
                             decommissioned. 2014-07-02
        :type decommission: string, date or datetime
        '''
        self.cm_ip = ip
        self.cm_name = name
        self.cm_desc = desc
        if type(decommission) in [date, datetime]:
            self.cm_decommission = decommission
        elif type(decommission) == str:
            # convert the string to datetime:
            self.cm_decommission = datetime.strptime(decommission, '%Y-%m-%d')

    @log_with(log)
    def store(self):
        Session.add(self)
        Session.commit()
        return True

    def to_json(self):
        return {'id': self.id,
                'ip': self.cm_ip,
                'name': self.cm_name,
                'desc': self.cm_desc,
                'decommission': str(self.cm_decommission)}
        

#
# The clienttoken table is the mapping which tokens belong to which client
#

machinetoken_table = sa.Table('MachineToken', meta.metadata,
                              sa.Column('id', sa.types.Integer(),
                                        sa.Sequence('machinetoken_seq_id',
                                                    optional=True),
                                        primary_key=True, nullable=False),
                              sa.Column("token_id", sa.types.Integer(),
                                        ForeignKey('Token.privacyIDEATokenId')),
                              sa.Column("machine_id", sa.types.Integer(),
                                        ForeignKey('ClientMachine.id')),
                              sa.Column("application", sa.types.Unicode(64)),
                              UniqueConstraint('token_id', 'machine_id',
                                               'application', name='uix_1'),
                              implicit_returning=implicit_returning,
                              )


class MachineToken(object):
    '''
    The MachineToken maps a token to a client and
    an application on this client
    
    The tuple of (machine, token, application) is unique.
    
    This can be an n:m mapping.
    '''

    @log_with(log)
    def __init__(self, machine_id, token_id, application):
        log.debug("setting machine_id to %r" % machine_id)
        self.machine_id = machine_id
        self.token_id = token_id
        self.application = application
        
    @log_with(log)
    def store(self):
        Session.add(self)
        Session.commit()
        return True
    
    def to_json(self):
        machinename = ""
        serial = ""
        ip = ""
        if self.token:
            serial = self.token.privacyIDEATokenSerialnumber
            is_active = self.token.privacyIDEAIsactive
        
        if self.machine:
            machinename = self.machine.cm_name
            ip = self.machine.cm_ip
        return {'id': self.id,
                'token_id': self.token_id,
                'serial': serial,
                'is_active': is_active,
                'machine_id': self.machine_id,
                'machinename': machinename,
                'ip': ip,
                'application': self.application}


machineuser_table = sa.Table('MachineUser', meta.metadata,
                             sa.Column('id', sa.types.Integer(),
                                       sa.Sequence('machineuser_seq_id',
                                                   optional=True),
                                       primary_key=True, nullable=False),
                             sa.Column('resolver', sa.types.Unicode(120),
                                       default=u'', index=True),
                             sa.Column('resclass', sa.types.Unicode(120),
                                       default=u''),
                             sa.Column('user_id', sa.types.Unicode(120),
                                       default=u'', index=True),
                             sa.Column("machine_id", sa.types.Integer(),
                                       ForeignKey('ClientMachine.id')),
                             sa.Column("application", sa.types.Unicode(64)),
                             UniqueConstraint('resolver', 'resclass',
                                              'user_id', 'machine_id',
                                              'application', name='uixu_1'),
                             implicit_returning=implicit_returning,
                             )
        
        
class MachineUser(object):
    '''
    The MachineUser maps a user to a client and
    an application on this client
    
    The tuple of (machine, USER, application) is unique.
    
    This can be an n:m mapping.
    '''

    @log_with(log)
    def __init__(self, machine_id,
                 resolver,
                 resclass,
                 user_id,
                 application):
        log.debug("setting machine_id to %r" % machine_id)
        self.machine_id = machine_id
        self.resolver = resolver
        self.resclass = resclass
        self.user_id = user_id
        self.application = application
        
    @log_with(log)
    def store(self):
        Session.add(self)
        Session.commit()
        return True
    
    def to_json(self):
        machinename = ""
        ip = ""
        if self.machine:
            machinename = self.machine.cm_name
            ip = self.machine.cm_ip
        return {'id': self.id,
                'user_id': self.user_id,
                'resolver': self.resolver,
                'resclass': self.resclass,
                'machine_id': self.machine_id,
                'machinename': machinename,
                'ip': ip,
                'application': self.application}


machine_t_options_table = sa.Table('MachineTokenOptions', meta.metadata,
                                   sa.Column('id', sa.types.Integer(),
                                             sa.Sequence('machinetokenoptions_seq_id', optional=True),
                                             primary_key=True, nullable=False),
                                   sa.Column('machinetoken_id',
                                             sa.types.Integer(),
                                             ForeignKey('MachineToken.id')),
                                   sa.Column('mt_key', sa.types.Unicode(64),
                                             nullable=False),
                                   sa.Column('mt_value', sa.types.Unicode(64),
                                             nullable=False),
                                   implicit_returning=implicit_returning)


class MachineTokenOptions(object):
    '''
    This class holds an Option for the token assigned to
    a certain client machine.
    Each Token-Clientmachine-Combination can have several
    options.
    '''
    
    def __init__(self, machinetoken_id, key, value):
        log.debug("setting %r to %r for MachineToken %s" % (key,
                                                            value,
                                                            machinetoken_id))
        self.machinetoken_id = machinetoken_id
        self.mt_key = key
        self.mt_value = value
        Session.add(self)
        Session.commit()


machine_u_options_table = sa.Table('MachineUserOptions', meta.metadata,
                                   sa.Column('id', sa.types.Integer(),
                                             sa.Sequence('machineuseroptions_seq_id', optional=True),
                                             primary_key=True,
                                             nullable=False),
                                   sa.Column('machineuser_id',
                                             sa.types.Integer(),
                                             ForeignKey('MachineUser.id')),
                                   sa.Column('mu_key', sa.types.Unicode(64),
                                             nullable=False),
                                   sa.Column('mu_value', sa.types.Unicode(64),
                                             nullable=False),
                                   implicit_returning=implicit_returning)


class MachineUserOptions(object):
    '''
    This class holds an Option for the Users assigned to
    a certain client machine.
    Each User-Clientmachine-Combination can have several
    options.
    '''
    
    def __init__(self, machineuser_id, key, value):
        log.debug("setting %r to %r for MachineUser %s" % (key,
                                                           value,
                                                           machineuser_id))
        self.machineuser_id = machineuser_id
        self.mu_key = key
        self.mu_value = value
        Session.add(self)
        Session.commit()


log.debug('calling ORM Mapper')


orm.mapper(Token, token_table, properties={
    'realms': relation(Realm, secondary=tokenrealm_table,
                       primaryjoin=token_table.c.privacyIDEATokenId == tokenrealm_table.c.token_id,
                       secondaryjoin=tokenrealm_table.c.realm_id == realm_table.c.id)
    })
orm.mapper(Realm, realm_table)
orm.mapper(TokenRealm, tokenrealm_table)
orm.mapper(Config, config_table)
orm.mapper(OcraChallenge, ocra_table)
orm.mapper(Challenge, challenges_table)

orm.mapper(Machine, clientmachine_table)
orm.mapper(MachineToken, machinetoken_table,
           properties={'token': relation(Token, 
                                         primaryjoin=token_table.c.privacyIDEATokenId == machinetoken_table.c.token_id),
                       'machine': relation(Machine,
                                           primaryjoin=clientmachine_table.c.id == machinetoken_table.c.machine_id),
                       'machine_t_options': relation(MachineTokenOptions,
                                                     primaryjoin=machinetoken_table.c.id == machine_t_options_table.c.machinetoken_id)
                       }) 
orm.mapper(MachineTokenOptions, machine_t_options_table)
# The MachineUser table is not connected with a non existing user table!
orm.mapper(MachineUser, machineuser_table,
           properties={'machine': relation(Machine,
                                           primaryjoin=clientmachine_table.c.id == machineuser_table.c.machine_id)})



# -*- coding: utf-8 -*-
#
#    privacyIDEA is a fork of LinOTP
#    May 28, 2014 Cornelius Kölbel
#    E-mail: info@privacyidea.org
#    Contact: www.privacyidea.org
#
#    Copyright (C) LinOTP: 2010 - 2014 LSE Leading Security Experts GmbH
#
#    This program is free software: you can redistribute it and/or
#    modify it under the terms of the GNU Affero General Public
#    License, version 3, as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the
#               GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#


class ISMSProvider:
    """ the SMS Provider Interface - BaseClass """
    def __init__(self):
        self.config = {}
        
    def submitMessage(self, phone, message):
        pass
    
    def loadConfig(self, configDict):
        self.config = configDict
        pass


def getSMSProviderClass(packageName, className):
    """ getSMSProviderClass(packageName, className):

    helper method to load the SMSProvider class from a given
    package in literal:
    
    example:
    
        getResolverClass("SkypeSMSProvider", "SMSProvider")()
    
    check:
        checks, if the submittMessage method exists
        if not an error is thrown
    
    """
    mod = __import__(packageName, globals(), locals(), [className])
    klass = getattr(mod, className)
    if not hasattr(klass, "submitMessage"):
        raise NameError("SMSProvider AttributeError: " + packageName + "." +
                        className + " instance of SMSProvider has no method"
                        " 'submitMessage'")
        return ""
    else:
        return klass


def main(phone, message):  # pragma: no cover
    print "SMSProvider - class load test "

    config = {'nothing': 'defined'}
    # sms = ISMSProvider()

    sms = getSMSProviderClass("SMSProvider", "ISMSProvider")()

    sms.loadConfig(config)
    sms.submitMessage(phone, message)
    print sms


if __name__ == "__main__":
    phone = "015154294800"
    message = "my test sms"
    main(phone, message)
    print "... done!"
    
    


# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#  HMAC-OTP (RFC 4226)
#  Copyright (C) LSE Leading Security Experts GmbH, Weiterstadt
#  Written by Max Vozeler <max.vozeler@lsexperts.de>
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
"""
  Description:  HOTP basic functions
"""
  
import hmac
import logging
import struct
import binascii

from hashlib import sha1
from privacyidea.lib.log import log_with

import sys
(ma, mi, _, _, _,) = sys.version_info
pver = float(int(ma) + int(mi) * 0.1)


log = logging.getLogger(__name__)


class HmacOtp():

    def __init__(self, secObj=None, counter=0, digits=6, hashfunc=sha1):
        self.secretObj = secObj
        self.counter = counter
        self.digits = digits
        self.hashfunc = hashfunc

    def hmac(self,
             counter=None,
             key=None,
             challenge=None):
        # log.error("hmacSecret()")
        counter = counter or self.counter

        # When using a counter, we can only use 64bit as data_input.
        # When we allow a raw data_input, we could use 160bit or more.
        if not challenge:
            data_input = struct.pack(">Q", counter)
        else:
            data_input = binascii.unhexlify(challenge)
            
        if key is None:
            dig = str(self.secretObj.hmac_digest(data_input, self.hashfunc))
        else:
            if pver > 2.6:
                dig = hmac.new(key, data_input, self.hashfunc).digest()
            else:
                dig = hmac.new(key, str(data_input), self.hashfunc).digest()

        return dig

    def truncate(self, digest):
        offset = ord(digest[-1:]) & 0x0f

        binary = (ord(digest[offset + 0]) & 0x7f) << 24
        binary |= (ord(digest[offset + 1]) & 0xff) << 16
        binary |= (ord(digest[offset + 2]) & 0xff) << 8
        binary |= (ord(digest[offset + 3]) & 0xff)

        return binary % (10 ** self.digits)

    def generate(self,
                 counter=None,
                 inc_counter=True,
                 key=None,
                 do_truncation=True,
                 challenge=None):
        
        counter = counter or self.counter

        if challenge:
            hmac = self.hmac(challenge=challenge, key=key)
        else:
            hmac = self.hmac(counter=counter, key=key)
        if do_truncation:
            otp = str(self.truncate(hmac))
            """  fill in the leading zeros  """
            sotp = (self.digits - len(otp)) * "0" + otp
        else:
            sotp = binascii.hexlify(hmac)
            
        if inc_counter:
            self.counter = counter + 1
        return sotp

    @log_with(log)
    def checkOtp(self, anOtpVal, window, symetric=False):
        res = -1
        start = self.counter
        end = self.counter + window
        if symetric is True:
            # changed window/2 to window for TOTP
            start = self.counter - (window)
            start = 0 if (start < 0) else start
            end = self.counter + (window)

        log.debug("OTP range counter: %r - %r" % (start, end))
        for c in range(start, end):
            otpval = self.generate(c)
            log.debug("calculating counter %r: %r %r"
                      % (c, anOtpVal, otpval))
            # log.error("otp[%d]: %s : %s",c,otpval,anOtpVal)

            if (unicode(otpval) == unicode(anOtpVal)):
                # log.debug("Match Pin: %s : %d : %s",otpval,c,anOtpVal)
                res = c
                break
        # return -1 or the counter
        return res

#eof##########################################################################


# -*- coding: utf-8 -*-
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2014-10-02 Activate SMS tests. PEP8
#             Cornelius Kölbel <cornelius@privacyidea.org>
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
functional tests
'''

import logging

import datetime

from privacyidea.tests import TestController, url


log = logging.getLogger(__name__)


class TestAuthorizeController(TestController):
    '''
    This test tests the authorization policies

    scope: authorization
    action: authorize
    realm:
    user:
    client:

        /validate/check
        /validate/simplecheck
        get_multi_otp
    '''

    def setUp(self):
        '''
        This sets up all the resolvers and realms
        '''
        TestController.setUp(self)
        self.curTime = datetime.datetime(2012, 5, 16, 9, 0, 52, 227413)
        self.TOTPcurTime = datetime.datetime.fromtimestamp(1337292860.585256)
        self.initToken()

    def createPWToken(self, serial, pin="", pw=""):
        '''
        creates the test tokens
        '''
        parameters = {"serial": serial,
                      "type": "PW",
                      # 64 byte key
                      "otpkey": pw,
                      "otppin": pin,
                      "otplen": len(pw),
                      "description": "PW testtoken",
                      }

        response = self.app.get(url(controller='admin', action='init'),
                                params=parameters)
        assert '"value": true' in response

    def removeTokenBySerial(self, serial):

        parameters = {"serial": serial}

        response = self.app.get(url(controller='admin', action='remove'),
                                params=parameters)
        return response

    def setTokenRealm(self, serial, realms):
        parameters = {"serial": serial,
                      "realms": realms}

        response = self.app.get(url(controller="admin", action="tokenrealm"),
                                params=parameters)
        return response

    def delPolicy(self, name):

        response = self.app.get(url(controller="system", action="delPolicy"),
                                params={"name": name})
        print "delPolicy:", response
        assert '"status": true' in response

    def setPolicy(self, parameters):
        response = self.app.get(url(controller='system', action='setPolicy'),
                                params=parameters)
        print "setPolicy:", response
        assert '"status": true' in response
        # check for policy
        response = self.app.get(url(controller='system', action='getPolicy'),
                                params=parameters)
        print "getPolicy:", response
        assert '"action": ' in response

    def initToken(self):
        '''
        init one DPW token
        '''
        self.createPWToken("pw1", pin="1234", pw="secret1")
        self.unassign("pw1")
        resp = self.app.get(url(controller='admin', action='assign'),
                            {'user': 'localuser',
                             'serial': 'pw1'})
        assert '"status": true' in resp
        resp = self.app.get(url(controller='admin', action='set'),
                            {'pin': '1234',
                             'serial': 'pw1'})
        assert '"status": true' in resp

        self.createPWToken("pw2", pin="1234", pw="secret2")
        self.unassign("pw2")
        resp = self.app.get(url(controller='admin', action='assign'),
                            {'user': 'horst',
                             'serial': 'pw2'})
        assert '"status": true' in resp
        resp = self.app.get(url(controller='admin', action='set'),
                            {'pin': '1234',
                             'serial': 'pw2'})
        assert '"status": true' in resp

    def create_policy(self):
        '''
        create the policy
        '''
        response = self.app.get(url(controller="system", action="setConfig"),
                                params={"mayOverwriteClient": None})
        print "setConfig:", response
        assert '"status": true' in response

        parameters1 = {'name': 'authorization1',
                       'scope': 'authorization',
                       'realm': '*',
                       'action': 'authorize',
                       'user': 'localuser',
                       'client': '172.16.200.0/24'
                       }
        self.setPolicy(parameters1)

    def create_policy2(self):
        '''
        create the policy
        '''
        response = self.app.get(url(controller="system", action="setConfig"),
                                params={"mayOverwriteClient": None})
        print "setConfig:", response
        assert '"status": true' in response

        parameters1 = {'name': 'authorization1',
                       'scope': 'authorization',
                       'realm': '*',
                       'action': 'authorize',
                       'user': 'horst',
                       'client': '172.16.200.0/24'
                       }
        self.setPolicy(parameters1)

    def create_policy3(self):
        '''
        create the policy
        '''
        response = self.app.get(url(controller="system", action="setConfig"),
                                params={"mayOverwriteClient": None})
        print "setConfig:", response
        assert '"status": true' in response

        parameters1 = {'name': 'authorization1',
                       'scope': 'authorization',
                       'realm': '*',
                       'action': 'authorize',
                       'user': '',
                       'client': '10.0.0.0/8'
                       }
        self.setPolicy(parameters1)

    def test_00_localuser_allowed(self):
        '''
        Auth Test 00: Without policy the user is authorized to login
        '''
        parameters = {'user': 'localuser',
                      'pass': '1234secret1',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

    def test_00_horst_allowed(self):
        '''
        Auth Test 00: without policy the user is allowed
        '''
        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

    def test_01_localuser_allowed(self):
        '''
        Auth Test 01: test if localuser is allowed to authenticate
        '''
        self.create_policy()

        parameters = {'user': 'localuser',
                      'pass': '1234secret1',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

    def test_02_horst_not_allowed(self):
        '''
        Auth Test 02: test if horst is not allowed to authenticate.
        horst is not authorized, since he is not mentioned in the
        policy1 as user.
        '''
        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": false' in response

    def test_03_localuser_not_allowed(self):
        '''
        Auth Test 03: localuser not allowed to authenticate to another host
        than 172.16.200.X
        localuser is not authorized, since he tries to login to 10.1.1.3
        '''
        parameters = {'user': 'localuser',
                      'pass': '1234secret1',
                      'client': '10.1.1.3'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": false' in response

    def test_04_horst_allowed(self):
        '''
        Auth Test 04: Now we set a new policy, and horst should be allowed
        '''
        self.create_policy2()

        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

    def test_05_blank_user(self):
        '''
        Auth Test 05: test if blank users are working for all users
        '''
        self.create_policy3()

        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'client': '10.0.1.2'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response


# ############################################################################
#
# PIN Policy tests
#
# There are two users in def-passwd:
#    localuser -> password test123
#    horst    -> password test123

    def test_06_pinpolicy(self):
        '''
        Auth Test 06: check a client policy with password PIN on one client
        '''

        # deleting authorization policy
        self.delPolicy("authorization1")

        # setting pin policy
        parameters = {'name': 'pinpolicy1',
                      'scope': 'authentication',
                      'realm': '*',
                      'action': 'otppin=1',
                      'user': '',
                      'client': '10.0.0.0/8'
                      }

        self.setPolicy(parameters)

        parameters = {'user': 'horst',
                      'pass': 'test123secret2',
                      'client': '10.0.1.2'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

    def test_07_pinpolicy(self):
        '''
        Auth Test 07: check on a client, that is not contained in policy
         => authenticate with OTP PIN
        '''
        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'client': '192.168.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

    def test_08_pinpolicy(self):
        '''
        Auth Test 08: check user on client, but user not contained in policy
        for this client => authenticate with OTP PIN
        '''
        parameters = {'name': 'pinpolicy2',
                      'scope': 'authentication',
                      'realm': '*',
                      'action': 'otppin=1',
                      'user': 'horst',
                      'client': '172.16.200.0/8'
                      }
        self.setPolicy(parameters)

        parameters = {'user': 'localuser',
                      'pass': '1234secret1',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

########################################################################
#
# Toke Type tests
#
    def test_10_tokentype(self):
        '''
        Auth Test 10: client not in policy. Every tokentype able to authenticate
        '''
        # clear pin policy
        self.delPolicy("pinpolicy1")
        self.delPolicy("pinpolicy2")
        
        parameters = {'name': 'tokentypepolicy1',
                      'scope': 'authorization',
                      'realm': '*',
                      'action': 'tokentype=HMAC',
                      'user': 'horst',
                      'client': '172.16.200.0/24'
                      }
        self.setPolicy(parameters)

        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'client': '10.0.0.2'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

    def test_11_tokentype(self):
        '''
        Auth Test 11: Tokentype policy contains list of tokentypes.
        A token is allowed to authenticate
        '''
        parameters = {'name': 'tokentypepolicy1',
                      'scope': 'authorization',
                      'realm': '*',
                      'action': 'tokentype=HMAC MOTP PW',
                      'user': 'horst',
                      'client': '172.16.200.0/24'
                      }
        self.setPolicy(parameters)

        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

    def test_12_tokentype(self):
        '''
        Auth Test 12: Tokentype policy contains list of tokentypes.
        the tokentype is not contained and not allowed to authenticate
        '''
        parameters = {'name': 'tokentypepolicy1',
                      'scope': 'authorization',
                      'realm': '*',
                      'action': 'tokentype=HMAC MOTP TOTP',
                      'user': 'horst',
                      'client': '172.16.200.0/24'
                      }
        self.setPolicy(parameters)

        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": false' in response

    def test_13_tokentype(self):
        '''
        Auth Test 13: Tokentype policy contains '*', the token type is allowed
        to authenticate.
        '''
        parameters = {'name': 'tokentypepolicy1',
                      'scope': 'authorization',
                      'realm': '*',
                      'action': 'tokentype=HMAC * TOTP MOTP',
                      'user': 'horst',
                      'client': '172.16.200.0/24'
                      }
        self.setPolicy(parameters)

        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='validate', action='check'),
                                params=parameters)

        print "validate/check: ", response
        assert '"value": true' in response

        self.delPolicy("tokentypepolicy1")

#####################################################################
#
# SMSText test
#
    def test_14_smstext(self):
        '''
        TODO: Testing policy for smstext
        '''
        pass

#####################################################################
#
#  AutoSMS

    def test_20_autosms(self):
        '''
        Auth Test 20: autosms enabled with no client and no user.
        Will do for all clients in a realm
        '''
        self.setPolicy({'name': 'autosms',
                        'scope': 'authentication',
                        'realm': '*',
                        'action': 'autosms',
                        'user': None})

        parameters = {'user': 'horst',
                      'client': '1.2.3.4'}
        response = self.app.get(url(controller='testing', action='autosms'),
                                params=parameters)

        print "testing/autosms: ", response
        assert '"value": true' in response

    def test_21_autosms(self):
        '''
        Auth Test 21: autosms enabled for client. autosms for client in subnet
        '''
        self.setPolicy({'name': 'autosms',
                        'scope': 'authentication',
                        'realm': '*',
                        'action': 'autosms',
                        'client': '172.16.200.0/24'
                        })

        parameters = {'user': 'horst',
                      'client': '172.16.200.123'}
        response = self.app.get(url(controller='testing', action='autosms'),
                                params=parameters)

        print "testing/autosms: ", response
        assert '"value": true' in response

    def test_22_autosms(self):
        '''
        Auth Test 22: Will not send autosms for a client outside this subnet
        '''
        self.setPolicy({'name': 'autosms',
                        'scope': 'authentication',
                        'realm': '*',
                        'action': 'autosms',
                        'client': '172.16.200.0/24'
                        })

        parameters = {'user': 'horst',
                      'client': '192.168.20.1'}
        response = self.app.get(url(controller='testing', action='autosms'),
                                params=parameters)

        print "testing/autosms: ", response
        assert '"value": false' in response

    def test_23_autosms(self):
        '''
        Auth Test 23: autosms enabled for a client and for a user.
        Will send autosms for this user
        '''
        self.setPolicy({'name': 'autosms',
                        'scope': 'authentication',
                        'realm': '*',
                        'action': 'autosms',
                        'client': '172.16.200.0/24',
                        'user': 'horst'
                        })

        parameters = {'user': 'horst',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='testing', action='autosms'),
                                params=parameters)

        print "testing/autosms: ", response
        assert '"value": true' in response

    def test_24_autosms(self):
        '''
        Auth Test 24: autosms enabled. Will not send autosms for another user
        '''
        self.setPolicy({'name': 'autosms',
                        'scope': 'authentication',
                        'realm': '*',
                        'action': 'autosms',
                        'client': '172.16.200.0/24',
                        'user': 'horst'
                        })

        parameters = {'user': 'localuser',
                      'client': '172.16.200.10'}
        response = self.app.get(url(controller='testing', action='autosms'),
                                params=parameters)

        print "testing/autosms: ", response
        assert '"value": false' in response

###################################################################
#
#   set realm tests
#
    def test_31_setrealm(self):
        '''
        Auth Test 31: setrealm for a user in the not default realm.
        '''
        self.delPolicy("tokentypepolicy1")

        self.setPolicy({"name": "setrealm",
                        "scope": "authorization",
                        "realm": "*",
                        "action": "setrealm=defRealm",
                        "client": "10.0.0.0/8"})
        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'realm': 'realm_does_not_exist',
                      'client': '10.0.0.1'}
        response = self.app.get(url(controller="validate", action="check"),
                                params=parameters)
        print "31 setrealm : ", response
        assert '"value": true'

    def test_32_setrealm(self):
        '''
        Auth Test 32: setrealm for a user, but user provides wrong password
        '''
        parameters = {'user': 'horst',
                      'pass': '1234secret2xxxx',
                      'realm': 'realm_does_not_exist',
                      'client': '10.0.0.1'}
        response = self.app.get(url(controller="validate", action="check"),
                                params=parameters)
        print "32 setrealm : ", response
        assert '"value": false'

    def test_33_setrealm(self):
        '''
        Auth Test 33: setrealm, but not for the right client
        '''
        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'realm': 'realm_does_not_exist',
                      'client': '172.0.0.1'}
        response = self.app.get(url(controller="validate", action="check"),
                                params=parameters)
        print "33 setrealm : ", response
        assert '"value": false'

    def test_34_setrealm(self):
        '''
        Auth Test 34: setrealm. rewrite realm to not existing realm. Auth fail
        '''
        self.setPolicy({"name": "setrealm",
                        "scope": "authorization",
                        "realm": "defRealm realm2 realm3",
                        "action": "setrealm=not_existing",
                        "client": "10.0.0.0/8"})

        parameters = {'user': 'horst',
                      'pass': '1234secret2',
                      'realm': 'defRealm',
                      'client': '10.0.0.1'}
        response = self.app.get(url(controller="validate", action="check"),
                                params=parameters)
        print "34 setrealm : ", response
        assert '"value": false'
        


# -*- coding: utf-8 -*-
#
#  Copyright (C) 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  cornelius@privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import yaml

from UserIdResolver import UserIdResolver
from UserIdResolver import getResolverClass

from sqlalchemy import and_
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from privacyidea.lib.phppass import PasswordHash
import traceback
import hashlib
from base64 import (b64decode,
                    b64encode)
import binascii

log = logging.getLogger(__name__)
ENCODING = "utf-8"

SQLSOUP_LOADED = False
try:
    from sqlsoup import SQLSoup
    SQLSOUP_LOADED = True
except:
    pass  # pragma: no cover

if SQLSOUP_LOADED is False:  # pragma: no cover
    try:
        from sqlalchemy.ext.sqlsoup import SQLSoup
        SQLSOUP_LOADED = True
    except:
        log.error("SQLSoup could not be loaded!")
        pass

       
class IdResolver (UserIdResolver):

    searchFields = {"username": "text",
                    "userid": "numeric",
                    "phone": "text",
                    "mobile": "text",
                    "surname": "text",
                    "givenname": "text",
                    "email": "text"
                    }

    @classmethod
    def setup(cls, config=None, cache_dir=None):
        '''
        this setup hook is triggered, when the server
        starts to serve the first request

        :param config: the privacyidea config
        :type  config: the privacyidea config dict
        '''
        log.info("Setting up the SQLResolver")
        return
    
    def __init__(self):
        self.resolverId = ""
        self.server = ""
        self.driver = ""
        self.database = ""
        self.port = 0
        self.limit = 100
        self.user = ""
        self.password = ""
        self.table = ""
        self.map = {}
        self.reverse_map = {}
        self.where = ""
        self.encoding = ""
        self.conParams = ""
        self.connect_string = ""
        self.session = None
        return

    def getSearchFields(self):
        return self.searchFields

    def checkPass(self, uid, password):
        """
        This function checks the password for a given uid.
        - returns true in case of success
        -         false if password does not match
        
        """
        def _check_ssha(pw_hash, password, hashfunc, length):
            pw_hash_bin = b64decode(pw_hash.split("}")[1])
            digest = pw_hash_bin[:length]
            salt = pw_hash_bin[length:]
            hr = hashfunc(password)
            hr.update(salt)
            return digest == hr.digest()
        
        def _check_sha(pw_hash, password):
            b64_db_password = pw_hash[5:]
            hr = hashlib.sha1(password).digest()
            b64_password = b64encode(hr)
            return b64_password == b64_db_password
        
        def _otrs_sha256(pw_hash, password):
            hr = hashlib.sha256(password)
            digest = binascii.hexlify(hr.digest())
            return pw_hash == digest
       
        res = False
        userinfo = self.getUserInfo(uid)
        
        database_pw = userinfo.get("password", "XXXXXXX")
        if database_pw[:2] == "$P":
            # We have a phpass (wordpress) password
            PH = PasswordHash()
            res = PH.check_password(password, userinfo.get("password"))
        # check salted hashed passwords
        elif database_pw[:6].upper() == "{SSHA}":
            res = _check_ssha(database_pw, password, hashlib.sha1, 20)
        elif database_pw[:9].upper() == "{SSHA256}":
            res = _check_ssha(database_pw, password, hashlib.sha256, 32)
        elif database_pw[:9].upper() == "{SSHA512}":
            res = _check_ssha(database_pw, password, hashlib.sha512, 64)
        # check for hashed password.
        elif userinfo.get("password", "XXXXX")[:5].upper() == "{SHA}":
            res = _check_sha(database_pw, password)
        elif len(userinfo.get("password")) == 64:
            # OTRS sha256 password
            res = _otrs_sha256(database_pw, password)
        
        return res
    
    def getUserInfo(self, userId):
        '''
        This function returns all user info for a given userid/object.
        
        :param userId: The userid of the object
        :type userId: string
        :return: A dictionary with the keys defined in self.map
        :rtype: dict
        '''
        userinfo = {}
        
        try:
            conditions = []
            column = self.map.get("userid")
            conditions.append(getattr(self.TABLE, column).like(userId))
            filter_condition = and_(*conditions)
            result = self.session.query(self.TABLE).filter(filter_condition)
                                                      
            for r in result:
                if len(userinfo.keys()) > 0:
                    raise Exception("More than one user with userid %s found!"
                                    % userId)
                userinfo = self._get_user_from_mapped_object(r)
        except Exception as exx:
            log.error("Could not get the userinformation: %r" % exx)
        
        return userinfo
    
    def getUsername(self, userId):
        '''
        Returns the username/loginname for a given userid
        :param userid: The userid in this resolver
        :type userid: string
        :return: username
        :rtype: string
        '''
        info = self.getUserInfo(userId)
        return info.get('username', "")
   
    def getUserId(self, LoginName):
        '''
        resolve the loginname to the userid.
        
        :param LoginName: The login name from the credentials
        :type LoginName: string
        :return: UserId as found for the LoginName
        '''
        userid = ""
        
        try:
            conditions = []
            column = self.map.get("username")
            conditions.append(getattr(self.TABLE, column).like(LoginName))
            filter_condition = and_(*conditions)
            result = self.session.query(self.TABLE).filter(filter_condition)
                                                      
            for r in result:
                if userid != "":
                    raise Exception("More than one user with loginname"
                                    " %s found!" % LoginName)
                user = self._get_user_from_mapped_object(r)
                userid = user["id"]
        except Exception as exx:
            log.error("Could not get the userinformation: %r" % exx)
        
        return userid
    
    def _get_user_from_mapped_object(self, ro):
        '''
        :param r: row
        :type r: Mapped Object
        :return: User
        :rtype: dict
        '''
        r = ro.__dict__
        user = {}
        try:
            if self.map.get("userid") in r:
                user["id"] = r[self.map.get("userid")]
        except UnicodeEncodeError:
            log.error("Failed to convert user: %r" % r)
            log.error(traceback.format_exc())
        
        for key in ["username",
                    "surname",
                    "givenname",
                    "email",
                    "mobile",
                    "phone",
                    "password"]:
            try:
                if r.get(self.map.get(key)):
                    val = r.get(self.map.get(key))
                    # val is a unicode!
                    # log.error("CKO: %r" % val)
                    # log.error("CKO: %r" % type(val))
                    user[key] = val
            except UnicodeEncodeError:
                log.error("Failed to convert user: %r" % r)
                log.error(traceback.format_exc())
        
        return user

    def getUserList(self, searchDict=None):
        '''
        :param searchDict: A dictionary with search parameters
        :type searchDict: dict
        :return: list of users, where each user is a dictionary
        '''
        users = []
        conditions = []
        if searchDict is None:
            searchDict = {}
        for key in searchDict.keys():
            column = self.map.get(key)
            value = searchDict.get(key)
            value = value.replace("*", "%")
            conditions.append(getattr(self.TABLE, column).like(value))
            
        if self.where:
            # this might result in erros if the
            # administrator enters nonsense
            (w_column, w_cond, w_value) = self.where.split()
            if w_cond.lower() == "like":
                conditions.append(getattr(self.TABLE, w_column).like(w_value))
            elif w_cond == "==":
                conditions.append(getattr(self.TABLE, w_column) == w_value)
            elif w_cond == ">":
                conditions.append(getattr(self.TABLE, w_column) > w_value)
            elif w_cond == "<":
                conditions.append(getattr(self.TABLE, w_column) < w_value)
        filter_condition = and_(*conditions)

        result = self.session.query(self.TABLE).\
            filter(filter_condition).\
            limit(self.limit)
                                                      
        for r in result:
            user = self._get_user_from_mapped_object(r)
            if "id" in user:
                users.append(user)
        return users
    
    def getResolverId(self):
        '''
        Returns the resolver Id
        This should be an Identifier of the resolver, preferable the type
        and the name of the resolver.
        '''
        return "sqlresolver." + self.resolverId

    @classmethod
    def getResolverClassType(cls):
        return 'sqlresolver'

    def getResolverType(self):
        return IdResolver.getResolverClassType()
    
    def loadConfig(self, config, conf):
        '''
        Load the config from conf.
        
        :param config: The configuration from the Config Table
        :type config: dict
        :param conf: the instance of the configuration
        :type conf: string
        
        The information which config entries we need to load is taken from
            manage.js: function save_sql_config
                    
        '''
        self.resolverId = conf
        self.server = self.getConfigEntry(config,
                                          'privacyidea.sqlresolver.Server',
                                          conf)
        self.driver = self.getConfigEntry(config,
                                          'privacyidea.sqlresolver.Driver',
                                          conf)
        self.database = self.getConfigEntry(config,
                                            'privacyidea.sqlresolver.Database',
                                            conf)
        self.port = self.getConfigEntry(config,
                                        'privacyidea.sqlresolver.Port',
                                        conf, required=False)
        self.limit = self.getConfigEntry(config,
                                         'privacyidea.sqlresolver.Limit',
                                         conf, required=False, default=100)
        self.user = self.getConfigEntry(config,
                                        'privacyidea.sqlresolver.User',
                                        conf, required=False)
        self.password = self.getConfigEntry(config,
                                            'privacyidea.sqlresolver.Password',
                                            conf, required=False)
        self.table = self.getConfigEntry(config,
                                         'privacyidea.sqlresolver.Table',
                                         conf)
        usermap = self.getConfigEntry(config,
                                      'privacyidea.sqlresolver.Map',
                                      conf)
        self.map = yaml.load(usermap)
        self.reverse_map = dict([[v, k] for k, v in self.map.items()])
        self.where = self.getConfigEntry(config,
                                         'privacyidea.sqlresolver.Where',
                                         conf, required=False)
        self.encoding = self.getConfigEntry(config,
                                            'privacyidea.sqlresolver.Encoding',
                                            conf,
                                            required=False, default="latin1")
        self.conParams = self.getConfigEntry(config,
                                             'privacyidea.sqlresolver.'
                                             'conParams',
                                             conf, required=False)
        
        # create the connectstring like
        params = {'Port': self.port,
                  'Password': self.password,
                  'conParams': self.conParams,
                  'Driver': self.driver,
                  'User': self.user,
                  'Server': self.server,
                  'Database': self.database}
        self.connect_string = self._create_connect_string(params)
        log.info("using the connect string %s" % self.connect_string)
        self.engine = create_engine(self.connect_string,
                                    encoding=self.encoding)
        # create a configured "Session" class
        Session = sessionmaker(bind=self.engine)

        # create a Session
        self.session = Session()
        self.db = SQLSoup(self.engine)
        self.TABLE = self.db.entity(self.table)
        
        return self

    def getResolverDescriptor(self):
        descriptor = {}
        typ = self.getResolverType()
        descriptor['clazz'] = "useridresolver.SQLIdResolver.IdResolver"
        descriptor['config'] = {'Server': 'string',
                                'Driver': 'string',
                                'Database': 'string',
                                'User': 'string',
                                'Password': 'string',
                                'Port': 'int',
                                'Limit': 'int',
                                'Table': 'string',
                                'Map': 'string',
                                'Where': 'string',
                                'Encoding': 'string',
                                'conParams': 'string'}
        return {typ: descriptor}

    def getConfigEntry(self, config, key, conf, required=True, default=None):
        ckey = key
        cval = ""
        if conf != "" or None:
            ckey = ckey + "." + conf
            if ckey in config:
                cval = config[ckey]
        if cval == "":
            if key in config:
                cval = config[key]
        if cval == "" and required is True:
            raise Exception("missing config entry: " + key)
        if cval == "" and default:
            cval = default
        return cval
            
    @classmethod
    def _create_connect_string(self, param):
        '''
        create the connectstring
        
        Port, Password, conParams, Driver, User,
        Server, Database
        '''
        port = ""
        password = ""
        conParams = ""
        if param.get("Port"):
            port = ":%s" % param.get("Port")
        if param.get("Password"):
            password = ":%s" % param.get("Password")
        if param.get("conParams"):
            conParams = "?%s" % param.get("conParams")
        connect_string = "%s://%s%s%s%s%s/%s%s" % (param.get("Driver", ""),
                                                   param.get("User", ""),
                                                   password,
                                                   "@" if (param.get("User")
                                                           or
                                                           password) else "",
                                                   param.get("Server", ""),
                                                   port,
                                                   param.get("Database", ""),
                                                   conParams)
        # SQLAlchemy does not like a unicode connect string!
        if param.get("Driver").lower() == "sqlite":
            connect_string = str(connect_string)
        log.debug("SQL connectstring: %r" % connect_string)
        return connect_string
            
    @classmethod
    def testconnection(self, param):
        '''
        This function lets you test the to be saved SQL connection.
              
        :param param: A dictionary with all necessary parameter
                        to test the connection.
        :type param: dict
        
        :return: Tuple of success and a description
        :rtype: (bool, string)
        
        Parameters are: Server, Driver, Database, User, Password, Port,
                        Limit, Table, Map
                        Where, Encoding, conParams
            
        '''
        num = -1
        desc = None
        
        connect_string = self._create_connect_string(param)
        log.info("using the connect string %s" % connect_string)
        engine = create_engine(connect_string)
        # create a configured "Session" class
        session = sessionmaker(bind=engine)()
        db = SQLSoup(engine)
        TABLE = db.entity(param.get("Table"))
            
        try:
            result = session.query(TABLE).count()
            num = result
            desc = "Found %i users." % num
        except Exception as exx:
            desc = "failed to retrieve users: %s" % exx
            
        return (num, desc)
    
    
if __name__ == "__main__":  # pragma: no cover

    print " SQLIdResolver - IdResolver class test "
        
    y = getResolverClass("SQLIdResolver", "IdResolver")()
    
    print y
    
    y.loadConfig({'privacyidea.sqlresolver.Driver': 'mysql',
                  'privacyidea.sqlresolver.Database': 'wordpress',
                  'privacyidea.sqlresolver.Server': 'localhost',
                  'privacyidea.sqlresolver.User': 'root',
                  'privacyidea.sqlresolver.Password': 'mspw.',
                  'privacyidea.sqlresolver.Limit': 2,
                  'privacyidea.sqlresolver.Encoding': "utf-8",
                  'privacyidea.sqlresolver.Table': 'wp_users',
                  'privacyidea.sqlresolver.Map': '{ "username": "user_login", \
                      "userid" : "ID", \
                      "email" : "user_email", \
                      "surname" : "display_name", \
                      "givenname" : "user_nicename", \
                      "password" : "user_pass"}',
                  },
                 "")

    print "Config loaded"
    
    print "====== getUserList =========="
    result = y.getUserList()
    for entry in result:
        print entry
    print "======== getUserId ==============="

    user = "admin"
    loginId = y.getUserId(user)

    print " %s -  %s" % (user, loginId)

    print " reId - " + y.getResolverId()
    print "======== getUserInfo ==============="

    ret = y.getUserInfo(loginId)
    print "Userinfo for %r" % loginId
    print ret
    print "============ checkPass =============="
    uid = y.getUserId("cornelius")
    ret = y.checkPass(uid, "test")
    print ret
    ret = y.checkPass(uid, "wrong")
    print ret
    print "============ getUserList ============"
    
    ret = y.getSearchFields()
    
    search = {"username": "admin"}
    
    print "=== we should only see the admin ===="
    ret = y.getUserList(search)
    for entry in ret:
        print entry



# -*- coding:utf-8 -*-
from mako import runtime, filters, cache
UNDEFINED = runtime.UNDEFINED
__M_dict_builtin = dict
__M_locals_builtin = locals
_magic_number = 9
_modified_time = 1422216484.849514
_enable_loop = True
_template_filename = '/usr/lib/python2.7/dist-packages/privacyidea/lib/tokens/yubicotoken.mako'
_template_uri = '/yubicotoken.mako'
_source_encoding = 'utf-8'
from webhelpers.html import escape
_exports = []


def render_body(context,**pageargs):
    __M_caller = context.caller_stack._push_frame()
    try:
        __M_locals = __M_dict_builtin(pageargs=pageargs)
        c = context.get('c', UNDEFINED)
        _ = context.get('_', UNDEFINED)
        __M_writer = context.writer()
        # SOURCE LINE 2
        __M_writer(u'\n\n')
        # SOURCE LINE 4
        if c.scope == 'config.title' :
            # SOURCE LINE 5
            __M_writer(u' ')
            __M_writer(escape(_("Yubico")))
            __M_writer(u'\n')
        # SOURCE LINE 7
        __M_writer(u'\n')
        # SOURCE LINE 8
        if c.scope == 'selfservice.title.enroll':
            # SOURCE LINE 9
            __M_writer(escape(_("Enroll Yubikey")))
            __M_writer(u'\n')
        # SOURCE LINE 11
        __M_writer(u'\n')
        # SOURCE LINE 12
        if c.scope == 'config' :
            # SOURCE LINE 13
            __M_writer(u'<script>\n\n/*\n * \'typ\'_get_config_val()\n *\n * this method is called, when the token config dialog is opened\n * - it contains the mapping of config entries to the form id\n * - according to the Config entries, the form entries will be filled\n *\n */\n\n\nfunction yubico_get_config_val(){\n\tvar id_map = {};\n\n    id_map[\'yubico.id\'] \t\t= \'sys_yubico_id\';\n    id_map[\'yubico.secret\'] = \'sys_yubico_secret\';\n\n\treturn id_map;\n\n}\n\n/*\n * \'typ\'_get_config_params()\n *\n * this method is called, when the token config is submitted\n * - it will return a hash of parameters for system/setConfig call\n *\n */\nfunction yubico_get_config_params(){\n\tvar url_params ={};\n\n    url_params[\'yubico.id\'] \t= $(\'#sys_yubico_id\').val();\n    url_params[\'yubico.secret\'] \t= $(\'#sys_yubico_secret\').val();\n\n\treturn url_params;\n}\n\n</script>\n\n<form class="cmxform" id=\'form_config_yubico\'>\n\t<p>\n\t\t')
            # SOURCE LINE 55
            __M_writer(escape(_("You get your own API key from the yubico website ")))
            __M_writer(u'\n\t\t<a href="https://upgrade.yubico.com/getapikey/" target="yubico">upgrade.yubico.com</a>.\n\t</p>\n\t<p>\n\t\t')
            # SOURCE LINE 59
            __M_writer(escape(_("If you do not use your own API key, the privacyIDEA demo API key will be used!")))
            __M_writer(u'\n\t</p>\n\t<table>\n\t<tr>\n\t<td><label for="sys_yubico_id" title=\'')
            # SOURCE LINE 63
            __M_writer(escape(_("You need to enter a valid API id")))
            __M_writer(u"'>\n\t\t")
            # SOURCE LINE 64
            __M_writer(escape(_("Yubico ID")))
            __M_writer(u'</label></td>\n\t<td><input class="required" type="text" name="sys_yubico_id" id="sys_yubico_id"\n\t\tclass="text ui-widget-content ui-corner-all"\n\t\tvalue="get-your-own"/></td>\n\t</tr>\n\n\t<tr>\n\t<td><label for="sys_yubico_secret" title=\'')
            # SOURCE LINE 71
            __M_writer(escape(_("You need to enter a valid API key")))
            __M_writer(u"'>\n\t\t")
            # SOURCE LINE 72
            __M_writer(escape(_("Yubico API key")))
            __M_writer(u'</label></td>\n\t<td><input class="required" type="text" name="sys_yubico_secret" id="sys_yubico_secret"\n\t\tclass="text ui-widget-content ui-corner-all"\n\t\tvalue="get-your-own"/></td>\n\t</tr>\n\n\t</table>\n\n</form>\n')
        # SOURCE LINE 82
        __M_writer(u'\n\n')
        # SOURCE LINE 84
        if c.scope == 'enroll.title' :
            # SOURCE LINE 85
            __M_writer(escape(_("Yubikey")))
            __M_writer(u'\n')
        # SOURCE LINE 87
        __M_writer(u'\n')
        # SOURCE LINE 88
        if c.scope == 'enroll' :
            # SOURCE LINE 89
            __M_writer(u'<script>\n\n/*\n * \'typ\'_get_enroll_params()\n *\n * this method is called, when the token  is submitted\n * - it will return a hash of parameters for admin/init call\n *\n */\n\nfunction yubico_get_enroll_params(){\n\tvar params ={};\n\n    params[\'yubico.tokenid\'] \t\t= $(\'#yubico_token_id\').val();\n    params[\'description\'] \t\t\t= $("#yubico_enroll_desc").val();\n\n\tjQuery.extend(params, add_user_data());\n\n\treturn params;\n}\n\n</script>\n\n<p>')
            # SOURCE LINE 112
            __M_writer(escape(_("Here you need to enter the token ID of the Yubikey.")))
            __M_writer(u'</p>\n<p>')
            # SOURCE LINE 113
            __M_writer(escape(_("You can do this by inserting the Yubikey and simply push the button.")))
            __M_writer(u'</p>\n<table>\n<tr>\n\t<td><label for="yubico_token_id" title=\'')
            # SOURCE LINE 116
            __M_writer(escape(_("You need to enter the Yubikey token ID")))
            __M_writer(u"'>\n\t\t")
            # SOURCE LINE 117
            __M_writer(escape(_("Token ID")))
            __M_writer(u'</label></td>\n\t<td><input class="required" type="text" name="yubico_token_id" id="yubico_token_id" min=12\n\t\tclass="text ui-widget-content ui-corner-all"/></td>\n</tr>\n<tr>\n    <td><label for="yubico_enroll_desc" id=\'yubico_enroll_desc_label\'>')
            # SOURCE LINE 122
            __M_writer(escape(_("Description")))
            __M_writer(u'</label></td>\n    <td><input type="text" name="yubico_enroll_desc" id="yubico_enroll_desc" value="Yubico Cloud token" class="text" /></td>\n</tr>\n</table>\n\n')
        # SOURCE LINE 128
        __M_writer(u'\n\n')
        # SOURCE LINE 130
        if c.scope == 'selfservice.enroll':
            # SOURCE LINE 131
            __M_writer(u'<script>\n\nfunction self_yubico_get_param()\n{\n\tvar urlparam = {};\n\tvar typ = \'yubico\';\n\n\turlparam[\'type\'] \t= typ;\n\turlparam[\'otplen\'] \t= 44;\n\turlparam[\'description\']    = $("#yubico_self_desc").val();\n\turlparam[\'yubico.tokenid\'] = $(\'#yubico_tokenid\').val();\n\n\treturn urlparam;\n}\n\nfunction self_tokenid_clear()\n{\n\t$(\'#yubico_tokenid\').val(\'\');\n\n}\nfunction self_yubico_submit(){\n\n\tvar ret = false;\n\tvar params =  self_yubico_get_param();\n\n\tenroll_token( params );\n\tret = true;\n\treturn ret;\n}\n\n</script>\n\n<h1>')
            # SOURCE LINE 163
            __M_writer(escape(_("Enroll your Yubikey")))
            __M_writer(u'</h1>\n<div id=\'enroll_yubico_form\'>\n\t<form class="cmxform" id=\'form_enroll_yubico\'>\n\t\t<p>\n\t\t\t')
            # SOURCE LINE 167
            __M_writer(escape(_("Enter the TokenId of your Yubikey. Simply insert the Yubikey and press the button.")))
            __M_writer(u"\n\t\t</p>\n\t<fieldset>\n\t\t<table>\n\t\t<tr>\n\t\t\t<td><label for='yubico_tokenid'>")
            # SOURCE LINE 172
            __M_writer(escape(_("Yubikey TokenId")+':'))
            __M_writer(u'</label></td>\n\t\t\t<td><input id=\'yubico_tokenid\' name=\'yubico_tokenid\'\n\t\t\t\tclass="required ui-widget-content ui-corner-all" min="12" maxlength=\'44\'/></td>\n\t\t</tr>\n\t\t<tr>\n\t\t    <td><label for="yubico_self_desc" id=\'yubico_self_desc_label\'>')
            # SOURCE LINE 177
            __M_writer(escape(_("Description")))
            __M_writer(u'</label></td>\n\t\t    <td><input type="text" name="yubico_self_desc" id="yubico_self_desc" value="Yubico Cloud (self)" class="text" /></td>\n\t\t</tr>\n\n        </table>\n\t    <button class=\'action-button\' id=\'button_enroll_yubico\'\n\t    \t    onclick="self_yubico_submit();">')
            # SOURCE LINE 183
            __M_writer(escape(_("enroll yubico token")))
            __M_writer(u'</button>\n\n    </fieldset>\n    </form>\n</div>\n\n')
        return ''
    finally:
        context.caller_stack._pop_frame()





# -*- coding: utf-8 -*-
#
#  privacyIDEA
#  June 30, 2014 Cornelius Kölbel, info@privacyidea.org
#  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
from privacyidea.model import Machine
from privacyidea.model import Token
from privacyidea.model import MachineToken
from privacyidea.model import MachineTokenOptions
from privacyidea.model.meta import Session
from privacyidea.lib.token import getTokens4UserOrSerial
from privacyidea.lib.token import getTokenType
from privacyidea.lib.applications import get_auth_item
from privacyidea.lib.applications import is_application_allow_bulk_call
from sqlalchemy import and_
from netaddr import IPAddress

import logging
log = logging.getLogger(__name__)
from privacyidea.lib.log import log_with


@log_with(log)
def create(name, ip=None, desc=None, decommission=None):
    machine = Machine(name, ip=ip, desc=desc, decommission=decommission)
    machine.store()
    log.info("Machine %r created." % machine)
    return machine


@log_with(log)
def delete(name):
    '''
    Delete the machine with the name and return the number of deleted machines
    
    Should always be 1
    Should be 0 if such a machine did not exist.
    '''
    mid = _get_machine_id(name)
    _assigns = Session.query(MachineToken)\
        .filter(MachineToken.machine_id == mid).delete()
    num = Session.query(Machine).filter(Machine.cm_name == name).delete()
    Session.commit()
    # 1 -> success
    return num == 1
    

@log_with(log)
def show(name=None, client_ip=None):
    '''
    Returns a dictionaries of machines matching name and client_ip
    
    :param name: the name of the machine
    :type name: string
    :param client_ip: the IP of the machine
    :type client_ip: sting
    :return: dictionary of machines, where the key is the machine name
             and the value is a dictionary with machine information.
    '''
    res = {}
    cond_list = []
    if name:
        cond_list.append(Machine.cm_name == name)
    if client_ip:
        cond_list.append(Machine.cm_ip == client_ip)
                  
    condition = and_(*cond_list)
    sqlquery = Session.query(Machine).filter(condition)
    for machine in sqlquery:
        res[machine.cm_name] = machine.to_json()
    return res


def _get_machine_id(machine_name, client_ip=None):
    # determine the machine_id for the machine name
    machine = show(machine_name, client_ip)
    if len(machine.keys()) > 1:
        raise Exception("Can not get ID for name=%r and IP=%r. "
                        "More than one machine found." %
                        (machine_name, client_ip))
    if len(machine.keys()) == 0:
        machine_id = None
    else:
        # There is only one machine in the dictionary and we get its ID
        machine_id = machine.values()[0].get("id")
        
    if machine_id is None:
        raise Exception("There is no machine with name=%r and IP=%r" %
                        (machine_name, client_ip))
    return machine_id


def _get_token_id(serial):
    # determine the token_id for the serial
    tokenlist = getTokens4UserOrSerial(serial=serial)
    if len(tokenlist) == 0:
        raise Exception("There is no token with the serial number %r" % serial)
    
    token_id = tokenlist[0].token.privacyIDEATokenId
    return token_id


def _get_machinetoken_id(machine_id, token_id, application):
    r = None
    sqlquery = Session.query(MachineToken.id)\
                .filter(and_(MachineToken.token_id == token_id,
                             MachineToken.machine_id == machine_id,
                             MachineToken.application == application))
    for row in sqlquery:
        r = row.id
    return r
    

def get_options(machine_id=None,
                token_id=None,
                application=None,
                machinetoken_id=None):
    """
    returns a dictionary of the options for a given tuple
    of machine, token and application from the table
    MachineTokenOptions.
    
    :param machine_id: id of the machine
    :param token_id: id ot the token
    :param application: name of the application
    :param machinetoken_id: id of the machineToken-entry
    
    :return: option dictionary
     
    You either need to specify (machine_ind, token_id, application) or
    the machinetoken_id.
    """
    options = {}
    if machinetoken_id:
        sqlquery = Session.query(MachineTokenOptions).\
            filter(MachineTokenOptions.machinetoken_id == machinetoken_id)
        for option in sqlquery:
            options[option.mt_key] = option.mt_value
    elif (machine_id and token_id and application):
        raise NotImplementedError("the tuple machine_id, token_id, "
                                  "application is not implemented, yet.")
    else:
        raise Exception("You either need to specify the machinetoken_id"
                        "or the tuple token_id, machine_id, application.")
    return options


def addoption(mtid=None,
              name=None,
              serial=None,
              application=None,
              options={}):
    """
    Add options to the machine token definition
    :param mtid: id of the machinetoken
    :param name: the machine name
    :param serial: the serial number of the token
    :param app: the application
    """
    if not mtid:
        mtid = _get_machinetoken_id(_get_machine_id(name),
                                    _get_token_id(serial),
                                    application)
    for option_name, option_value in options.items():
            MachineTokenOptions(mtid, option_name, option_value)
    return len(options)


def deloption(mtid=None,
              name=None,
              serial=None,
              application=None,
              key=None):
    """
    delete option from a machine token definition
    :param mtid: id of the machinetoken
    :param name: the machine name
    :param serial: the serial number of the token
    :param app: the application
    """
    if not mtid:
        mtid = _get_machinetoken_id(_get_machine_id(name),
                                    _get_token_id(serial),
                                    application)
    num = Session.query(MachineTokenOptions).\
                        filter(and_(MachineTokenOptions.machinetoken_id == mtid,
                                    MachineTokenOptions.mt_key == key)).delete()
    Session.commit()
    return num == 1


@log_with(log)
def addtoken(machine_name,
             serial,
             application,
             options=None):
    machine_id = _get_machine_id(machine_name)
    if not machine_id:
        raise Exception("No machine with name %r found!" % machine_name)
    token_id = _get_token_id(serial)
    if not token_id:
        raise Exception("No token with serial %r found!" % serial)
    machinetoken = MachineToken(machine_id, token_id, application)
    machinetoken.store()
    # Add options to the machine token
    if options:
        addoption(machinetoken.id, options=options)
            
    return machinetoken

  
@log_with(log)
def deltoken(machine_name, serial, application):
    """
    Delete a machine token.
    Also deletes the corresponding MachineTokenOptions
    """
    machine_id = _get_machine_id(machine_name)
    token_id = _get_token_id(serial)
    mtid = _get_machinetoken_id(machine_id, token_id, application)
    
    Session.query(MachineTokenOptions).\
        filter(MachineTokenOptions.machinetoken_id == mtid).delete()
    num = Session.query(MachineToken).\
        filter(and_(MachineToken.token_id == token_id,
                    MachineToken.machine_id == machine_id,
                    MachineToken.application == application)).delete()
    
    Session.commit()
    # 1 -> success
    return num == 1
    
    
@log_with(log)
def showtoken(machine_name=None,
              serial=None,
              application=None,
              client_ip=None,
              flexi=None,
              params=None):
    '''
    Returns a dictionary of tokens assigned to machines.
    
    :param flexi: If set, the output will be in flexigrid format and
            we will output all machines, even if they have no token assigned
    :type flexi: boolean
    :return: JSON of all tokens connected to machines with the corresponding
             application.
    '''
    res = {}
    machine_id = None
    token_id = None
    cond_list = []
    # default
    order = Machine.cm_name
    page = 1
    page_size = 15

    if machine_name or client_ip:
        machine_id = _get_machine_id(machine_name, client_ip)
    if serial:
        token_id = _get_token_id(serial)
    
    if machine_id:
        cond_list.append(MachineToken.machine_id == machine_id)
    if token_id:
        cond_list.append(MachineToken.token_id == token_id)
    if application:
        cond_list.append(MachineToken.application == application)
    
    machines = {}
     
    # For flexigrid we use addtional parameter
    if flexi:
        if params:
            # Filtering
            qtype = params.get("qtype")
            query = params.get("query")
            if qtype == "machine":
                cond_list.append(Machine.cm_name.like("%" + query + "%"))
            elif qtype == "IP":
                cond_list.append(Machine.cm_ip.like("%" + query + "%"))
            elif qtype == "description":
                cond_list.append(Machine.cm_desc.like("%" + query + "%"))
            elif qtype == "serial":
                cond_list.append(Token.privacyIDEATokenSerialnumber.
                                 like("%" + query + "%"))
            elif qtype == "application":
                cond_list.append(MachineToken.application.like("%" 
                                                               + query + "%"))

            condition = and_(*cond_list)

            # Flexigrid sorting
            sort = params.get("sortname")
            sort_order = params.get("sortorder")
            if sort == "machine_id":
                order = MachineToken.machine_id
            elif sort == "machine":
                order = Machine.cm_name
            elif sort == "IP":
                order = Machine.cm_ip
            elif sort == "description":
                order = Machine.cm_desc
            elif sort == "serial":
                order = Token.privacyIDEATokenSerialnumber
            elif sort == "application":
                order = MachineToken.application
            
            if sort_order is not None and sort_order == "desc":
                order = order.desc()
            else:
                order = order.asc()
                
            # pagination
            page = int(params.get("page", 1))
            page_size = int(params.get("rp", 15))

        sqlquery = Session.query(Machine,
                                 MachineToken.id,
                                 MachineToken.application,
                                 Token.privacyIDEATokenSerialnumber,
                                 Token.privacyIDEAIsactive)\
                                        .outerjoin(MachineToken)\
                                        .outerjoin(Token)\
                                        .filter(condition)\
                                        .order_by(order)\
                                        .limit(page_size)\
                                        .offset(page_size*(page - 1))
        # Fixme: This is not best way to determine the total count.
        sql_total = Session.query(Machine,
                                  MachineToken.application,
                                  Token.privacyIDEATokenSerialnumber)\
                                        .outerjoin(MachineToken)\
                                        .outerjoin(Token)\
                                        .filter(condition)
        total = 0
        for _i in sql_total:
            total += 1
        rows = []
        m_id = 0
        for row in sqlquery:
            machine, _mtid, application, serial, is_active = row
            m_id += 1
            rows.append({'id': m_id,
                         'cell': [(m_id),
                                  (machine.id),
                                  (machine.cm_name),
                                  (machine.cm_ip),
                                  (machine.cm_desc),
                                  (serial) or "",
                                  (is_active),
                                  (application) or ""]})
        
        res = {"page": page,
               "total": total}
        res["rows"] = rows
    else:
        condition = and_(*cond_list)
        sqlquery = Session.query(MachineToken).filter(condition)
        for row in sqlquery:
            machines[row.id] = row.to_json()
            # add options
            if application:
                machine_options = get_options(machinetoken_id=row.id)
                machines[row.id]["options"] = machine_options

        res["total"] = len(machines)
        res["machines"] = machines

    Session.commit()
    return res


@log_with(log)
def get_token_apps(machine=None,
                   application=None,
                   application_module=None,
                   serial=None,
                   client_ip=None,
                   challenge=None):
    '''
    This method returns the authentication data for the
    requested application and token
    
    :param machine: the machine name (optional)
    :param application: the name of the application (optional)
    :param client: the IP of the client (required)
    :param serial: the serial number of a specific token (optional)
    :param challenge: a challenge parameter, that can be passed in selfTest
    '''
    if not client_ip:
        log.warning("No client IP.")
        return {}
    if not IPAddress(client_ip):
        log.warning("No valid client IP: %r" % client_ip)
        return {}

    # if the application has allow_bulk_action set, we need to
    # remove the IP filer.
    if application_module:
        if is_application_allow_bulk_call(application_module):
            client_ip = None
    res = showtoken(machine_name=machine,
                    client_ip=client_ip,
                    application=application,
                    serial=serial)
    '''
    depending on the application type we do need to take some action
    Each application should know, what to provide...
    Determine this by
     1. application
     2. token type
     3. serial number
    '''
    machines = res.get("machines")
    if application:
        for machine in machines.values():
            # add token information
            serial = machine.get("serial")
            token_type = getTokenType(serial)
            auth_item = get_auth_item(application,
                                      application_module,
                                      token_type,
                                      serial,
                                      challenge=challenge)
            machine["auth_item"] = auth_item
            # add options
            machine_options = get_options(machinetoken_id=machine.get("id"))
            machine["options"] = machine_options
  
    return res



# -*- coding: utf-8 -*-
#
#    privacyIDEA Feitian ChalRep test suite
#
#    Copyright (C)  2014 Cornelius Kölbel, cornelius@privacyidea.org
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
log = logging.getLogger(__name__)
import unittest
from privacyidea.lib.feitian import (calculate_optical_data,
                                     create_image,
                                     ACCOUNT,
                                     AMOUNT)


class TestFeitianChalRep(unittest.TestCase):

    def setUp(self):
        pass
    
    def test_feitian_chal_12345678(self):
        # data for challenge 12345678
        demo_data_challenge = [0, 1, 2, 0, 1, 2, 1, 0, 1, 0, 2, 1, 0, 1, 0, 2,
                               1, 0, 2, 1, 2, 1, 0, 1, 0, 2, 1, 2, 0, 2, 0, 2,
                               1, 0, 1, 0, 1, 0, 1, 2, 1, 0, 1, 2, 0, 1, 0, 2,
                               1, 0, 2, 0]
        optical = calculate_optical_data(value="12345678")
        print "-" * 70
        print "Cha D: %s" % demo_data_challenge
        print "Cha O: %s " % optical
        create_image(optical)
        assert demo_data_challenge == optical
    
    def test_feitian_chal_12345(self):
        # data for account 12345
        demo_data_account = [0, 1, 2, 0, 1, 2, 1, 2, 0, 2, 1, 0, 2, 1, 2, 1, 2,
                             1, 0, 2, 0, 2, 1, 2, 1, 0, 2, 0, 1, 0, 1, 0, 2, 1,
                             2, 1, 2, 1, 2, 0]
        optical = calculate_optical_data(value="12345", type=ACCOUNT)
        print "-" * 70
        print "Acc D: %s" % demo_data_account
        print "Acc O: %s " % optical
        create_image(optical, outfile="account.gif")
        assert demo_data_account == optical
    
    def test_feitian_chal_amount(self):
        # data for amount
        demo_data_amount = [0, 1, 2, 0, 1, 2, 1, 2, 1, 0, 2, 1, 0, 2, 0, 2, 0,
                            2, 0, 1, 0, 2, 0, 1, 2, 0, 2, 1, 0, 1, 0, 2, 0, 2,
                            1, 0, 2, 1, 2, 0]
        optical = calculate_optical_data(value="67890", type=AMOUNT)
        print "-" * 70
        print "Amo D: %s" % demo_data_amount
        print "Amo O: %s " % optical
        create_image(optical, outfile="amount.gif")
        assert demo_data_amount == optical
    
    
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  Aug 15, 2014 Cornelius Kölbel: CHange login behaviour
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging
log = logging.getLogger(__name__)

from privacyidea.lib.util import getRealmBox
from privacyidea.lib.realm import getDefaultRealm
from privacyidea.lib.selftest import isSelfTest
import traceback
from privacyidea.lib.account import check_admin_password
from privacyidea.lib.account import authenticate_privacyidea_user
from privacyidea.lib.account import is_admin_identity
from privacyidea.lib.user import check_user_password
from privacyidea.lib.user import splitUser
from privacyidea.lib.log import log_with

from pylons import request, config, tmpl_context as c
from privacyidea.lib.config import get_privacyIDEA_config
from privacyidea.lib.config import getFromConfig
from privacyidea.lib.policy import PolicyClass
ENCODING = "utf-8"


class UserModelPlugin(object):

    def _get_user_from_login(self, login, default_realm=True):
        '''
        take the login and return a username and a realm name
        :default_realm: If set to True and no realm is found, the
                        default realm will be returned
        :return: tuple (username, realm)
        '''
        username = login
        realm = ""
        
        splitAtSign = getFromConfig("splitAtSign", "true")
        if splitAtSign.lower() == "true":
            (username, realm) = splitUser(login)

        if realm == "" and default_realm:
                realm = getDefaultRealm()
            
        return (username, realm)

    @log_with(log)
    def authenticate(self, environ, identity):
        username = None
        realm = None
        success = None
        try:
            if isSelfTest():
                if ('login' not in identity and 'repoze.who.plugins.auth_tkt.userid' in identity):
                    u = identity.get('repoze.who.plugins.auth_tkt.userid')
                    identity['login'] = u
                    identity['password'] = u

            username, realm = self._get_user_from_login(identity['login'],
                                                        default_realm=False)
            if getRealmBox() and realm == "":
                # The login name contained no realm
                realm = identity['realm']
            if realm == "":
                # The realm is still empty
                realm = getDefaultRealm()

            password = identity['password']
        except KeyError as e:
            log.error("Keyerror in identity: %r." % e)
            log.error("%s" % traceback.format_exc())
            return None

        # check username/realm, password
        if isSelfTest():
            success = "%s@%s" % (unicode(username), unicode(realm))
        else:
            Policy = PolicyClass(request, config, c,
                                 get_privacyIDEA_config())
            if Policy.is_auth_selfservice_otp(username, realm):
                # check the OTP
                success = authenticate_privacyidea_user(username,
                                                        realm,
                                                        password)
            else:
                # We do authentication against the user store
                success = check_user_password(username, realm, password)

        if not success and is_admin_identity("%s@%s" % (username, realm),
                                             exception=False):
            # user not found or authenticated in resolver.
            # So let's see, if this is an administrative user.
            success = check_admin_password(username, password, realm)

        if success:
            log.info("User %r authenticated" % success)
        return success

    @log_with(log)
    def add_metadata(self, environ, identity):
        # username = identity.get('repoze.who.userid')
        # user = User.get(username)
        # user = "conelius koelbel"
        # log.info( "add_metadata: %s" % identity )

        # pp = pprint.PrettyPrinter(indent=4)

        # if identity.has_key('realm'):
        #    identity.update( { 'realm' : identity['realm'] } )

        return identity



# -*- coding: utf-8 -*-
import codecs

from pyparsing import White, Word, alphanums, CharsNotIn
from pyparsing import Forward, Group, OneOrMore
from pyparsing import pythonStyleComment
from privacyidea.lib.freeradiusparser import BaseParser


class CronJobParser(BaseParser):
    
    dtime = Word("0123456789-*")
    command = CharsNotIn("{}\n#,")
    username = Word(alphanums)
    key = Word(alphanums)
    value = command
    space = White().suppress()
    comment = Word("#")
    equals = Word("=")
    assignment = Forward()
    assignment << Group(key
                        + equals.suppress()
                        + value)
    entry_block = Forward()
    entry_block << Group(dtime
                         + space.suppress()
                         + dtime
                         + space.suppress()
                         + dtime
                         + space.suppress()
                         + dtime
                         + space.suppress()
                         + dtime
                         + space.suppress()
                         + username
                         + space.suppress()
                         + command
                         )
    cron_file = OneOrMore(entry_block | assignment).ignore(pythonStyleComment)
    
    file_header = """# File parsed and saved by privacyidea.\n\n"""
    
    def __init__(self,
                 infile="/etc/crontab",
                 content=None):
        self.file = None
        if content:
            self.content = content
        else:
            self.file = infile
            f = codecs.open(self.file, "r", "utf-8")
            self.content = f.read()
            f.close()
            
    def get(self):
        """
        return the grouped config
        
        """
        # reread the file from dist
        if self.file:
            f = codecs.open(self.file, "r", "utf-8")
            self.content = f.read()
            f.close()
        config = self.cron_file.parseString(self.content)
        return config
    
    def format(self, config):
        '''
        :return: The formatted data as it would be written to a file
        '''
        output = ""
        output += self.file_header
        # write the assignments
        assignments = config.get("assignments")
        for assignment in assignments:
            output += "%s=%s\n" % (assignment,
                                   assignments.get(assignment))
        # write the cronjobs
        output += "\n#m\th\tdom\tmon\tdow\tuser\tcommand\n"
        cronjobs = config.get("cronjobs")
        for entry in cronjobs:
            output += "%s\t%s\t%s\t%s\t%s\t%s\t%s" % (entry.get("minute", "*"),
                                                      entry.get("hour", "*"),
                                                      entry.get("dom", "*"),
                                                      entry.get("month", "*"),
                                                      entry.get("dow", "*"),
                                                      entry.get("user"),
                                                      entry.get("command"))
            output += "\n"
        return output
    
    def get_dict(self):
        data = self.get()
        assignments = {}
        cronjobs = []
        for entry in data:
            if len(entry) == 2:
                assignments[entry[0]] = entry[1]
            elif len(entry) == 7:
                cronjobs.append({"minute": entry[0],
                                 "hour": entry[1],
                                 "dom": entry[2],
                                 "month": entry[3],
                                 "dow": entry[4],
                                 "user": entry[5],
                                 "command": entry[6]
                                 })
        
        return {"assignments": assignments,
                "cronjobs": cronjobs}


def main():  # pragma: no cover
    CP = CronJobParser()
    config = CP.get_dict()
    print CP.format(config)
        

if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
#
#  product:  privacyIDEA is a fork of LinOTP
#  May, 08 2014 Cornelius Kölbel
#  http://www.privacyidea.org
#
#  2014-10-19 Remove class SecurityModule from __init__.py
#             and add it here.
#             Cornelius Kölbel <cornelius@privacyidea.org>
#
#  product:  LinOTP2
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import logging
import binascii
import os

from Crypto.Cipher import AES

from privacyidea.lib.crypto import zerome

TOKEN_KEY = 0
CONFIG_KEY = 1
VALUE_KEY = 2
DEFAULT_KEY = 3


log = logging.getLogger(__name__)


class SecurityModule(object):

    def __init__(self, config=None):
        log.error("This is the base class. You should implement this!")
        self.name = "SecurityModule"

    def isReady(self):
        fname = 'isReady'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented %s"
                                  % fname)

    def setup_module(self, params):
        fname = 'setup_module'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented %s"
                                  % fname)

    ''' base methods '''
    def random(self, length):
        fname = 'random'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented %s"
                                  % fname)

    def encrypt(self, value, iv=None):
        fname = 'encrypt'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented %s"
                                  % fname)

    def decrypt(self, value, iv=None):
        fname = 'decrypt'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented %s"
                                  % fname)

    ''' higer level methods '''
    def encryptPassword(self, cryptPass):
        fname = 'decrypt'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented %s"
                                  % fname)

    def encryptPin(self, cryptPin):
        fname = 'decrypt'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented %s"
                                  % fname)

    def decryptPassword(self, cryptPass):
        fname = 'decrypt'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented %s"
                                  % fname)

    def decryptPin(self, cryptPin):
        fname = 'decrypt'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented %s"
                                  % fname)


class DefaultSecurityModule(SecurityModule):

    def __init__(self, config=None):
        '''
        initialsation of the security module

        :param config:  contains the configuration definition
        :type  config:  - dict -

        :return -
        '''

        self.name = "Default"
        self.config = config
        self.crypted = False
        self.is_ready = True
        self._id = binascii.hexlify(os.urandom(3))

        if "crypted" in config:
            crypt = config.get('crypted').lower()
            if crypt == 'true':
                self.crypted = True
                self.is_ready = False

        if "file" in config is False:
            log.error("No secret file defined. A parameter "
                      "privacyideaSecretFile is missing in your "
                      "privacyidea.ini.")
            raise Exception("no secret file defined: privacyideaSecretFile!")

        self.secFile = config.get('file')
        self.secrets = {}

        return

    def isReady(self):
        '''
        provides the status, if the security module is fully initializes
        this is required especially for the runtime confi like set password ++

        :return:  status, if the module is fully operational
        :rtype:   boolean

        '''
        return self.is_ready

    def getSecret(self, slot_id=0):
        '''
        internal function, which accesses the key in the defined slot

        :param slot_id: slot id of the key array
        :type  slot_id: int

        :return: key or secret
        :rtype:  binary string

        '''
        log.debug('getSecret()')
        slot_id = int(slot_id)

        if self.crypted:
            if slot_id in self.secrets:
                return self.secrets.get(slot_id)

        secret = ''
        try:
                f = open(self.secFile)
                for _i in range(0, slot_id + 1):
                    secret = f.read(32)
                f.close()
                if secret == "":
                    # secret = setupKeyFile(secFile, slot_id+1)
                    raise Exception("No secret key defined for index: %s !\n"
                                    "Please extend your %s"" !"
                                    % (str(id), self.secFile))
        except Exception as e:
            raise Exception("Exception:" + unicode(e))

        if self.crypted:
            self.secrets[id] = secret

        return secret

    def setup_module(self, param):
        '''
        callback, which is called during the runtime to initialze the
        security module

        :param params: all parameters, which are provided by the http request
        :type  params: dict

        :return: -

        '''
        if self.crypted is False:
            return
        if "password" in param is False:
            raise Exception("missing password")

        # if we have a crypted file and a password, we take all keys
        # from the file and put them in a hash
        #
        # After this we do not require the password anymore

        handles = ['pinHandle', 'passHandle', 'valueHandle', 'defaultHandle']
        for handle in handles:
            self.getSecret(self.config.get(handle, '0'))

        self.is_ready = True
        return

    # the real interfaces: random, encrypt, decrypt '''
    def random(self, length=32):
        '''
        security module methods: random

        :param len: length of the random byte array
        :type  len: int

        :return: random bytes
        :rtype:  byte string
        '''

        log.debug('random()')
        return os.urandom(length)

    def encrypt(self, data, iv, slot_id=0):
        '''
        security module methods: encrypt

        :param data: the to be encrypted data
        :type  data:byte string

        :param iv: initialisation vector (salt)
        :type  iv: random bytes

        :param  slot_id: slot of the key array
        :type   slot_id: int

        :return: encrypted data
        :rtype:  byte string
        '''

        log.debug('encrypt()')

        if self.is_ready is False:
            raise Exception('setup of security module incomplete')

        key = self.getSecret(slot_id)
        # convert input to ascii, so we can securely append bin data
        input_data = binascii.b2a_hex(data)
        input_data += u"\x01\x02"
        padding = (16 - len(input_data) % 16) % 16
        input_data += padding * "\0"
        aes = AES.new(key, AES.MODE_CBC, iv)

        # cko: ARGH: Only ECB!
        # import privacyideaee.lib.yhsm as yhsm
        # y = yhsm.YubiHSM(0x1111, password="14fda9321ae820aa34e57852a31b10d0")
        # y.unlock(password="14fda9321ae820aa34e57852a31b10d0")
        # res = y.encrypt(input)
        #
        res = aes.encrypt(input_data)

        if self.crypted is False:
            zerome(key)
            del key
        return res

    def decrypt(self, input_data, iv, slot_id=0):
        '''
        security module methods: decrypt

        :param input_data: the to be decrypted data
        :type  input_data: byte string

        :param iv: initialisation vector (salt)
        :type  iv: random bytes

        :param  slot_id: slot of the key array
        :type   slot_id: int

        :return: decrypted data
        :rtype:  byte string
        '''

        log.debug('decrypt()')

        if self.is_ready is False:
            raise Exception('setup of security module incomplete')

        key = self.getSecret(slot_id)
        aes = AES.new(key, AES.MODE_CBC, iv)
        # cko
        # import privacyideaee.lib.yhsm as yhsm
        # y = yhsm.YubiHSM(0x1111, password="14fda9321ae820aa34e57852a31b10d0")
        # y.unlock(password="14fda9321ae820aa34e57852a31b10d0")
        # log.debug("CKO in: %s" % input)
        # output = binascii.hexlify(y.decrypt(input))
        # log.debug("CKO out: %s" % output)
        #
        output = aes.decrypt(input_data)
        # log.debug("CKO: output2: %s" % output)
        eof = output.rfind(u"\x01\x02")
        if eof >= 0:
            output = output[:eof]

        # convert output from ascii, back to bin data
        data = binascii.a2b_hex(output)

        if self.crypted is False:
            zerome(key)
            del key

        return data

    def decryptPassword(self, cryptPass):
        '''
        dedicated security module methods: decryptPassword
        which used one slot id to decryt a string

        :param cryptPassword: the crypted password - leading iv, seperated
                            by the ':'
        :param cryptPassword: byte string

        :return: decrypted data
        :rtype:  byte string
        '''

        return self._decryptValue(cryptPass, CONFIG_KEY)

    def decryptPin(self, cryptPin):
        '''
        dedicated security module methods: decryptPin
        which used one slot id to decryt a string

        :param cryptPin: the crypted pin - - leading iv, seperated by the ':'
        :param cryptPin: byte string

        :return: decrypted data
        :rtype:  byte string
        '''

        return self._decryptValue(cryptPin, TOKEN_KEY)

    def encryptPassword(self, password):
        '''
        dedicated security module methods: encryptPassword
        which used one slot id to encrypt a string

        :param password: the to be encrypted password
        :param password: byte string

        :return: encrypted data - leading iv, seperated by the ':'
        :rtype:  byte string
        '''
        return self._encryptValue(password, CONFIG_KEY)

    def encryptPin(self, pin):
        '''
        dedicated security module methods: encryptPin
        which used one slot id to encrypt a string

        :param pin: the to be encrypted pin
        :param pin: byte string

        :return: encrypted data - leading iv, seperated by the ':'
        :rtype:  byte string
        '''
        return self._encryptValue(pin, TOKEN_KEY)

    ''' base methods for pin and password '''
    def _encryptValue(self, value, slot_id):
        '''
        _encryptValue - base method to encrypt a value
        - uses one slot id to encrypt a string
        retrurns as string with leading iv, seperated by ':'

        :param value: the to be encrypted value
        :param value: byte string

        :param  slot_id: slot of the key array
        :type   slot_id: int

        :return: encrypted data with leading iv and sepeartor ':'
        :rtype:  byte string
        '''
        iv = self.random(16)
        v = self.encrypt(value, iv, slot_id)

        value = binascii.hexlify(iv) + ':' + binascii.hexlify(v)
        return value

    def _decryptValue(self, cryptValue, slot_id):
        '''
        _decryptValue - base method to decrypt a value
        - used one slot id to encrypt a string with leading iv,
        seperated by ':'

        :param cryptValue: the to be encrypted value
        :param cryptValue: byte string

        :param  slot_id: slot of the key array
        :type   slot_id: int

        :return: decrypted data
        :rtype:  byte string
        '''
        # split at ":"
        pos = cryptValue.find(':')
        bIV = cryptValue[:pos]
        bData = cryptValue[pos + 1:len(cryptValue)]

        iv = binascii.unhexlify(bIV)
        data = binascii.unhexlify(bData)

        password = self.decrypt(data, iv, slot_id)

        return password


class ErrSecurityModule(DefaultSecurityModule):

        def setup_module(self, params):
            ret = DefaultSecurityModule.setup_module(self, params)
            self.is_ready = False
            return ret


# eof#######################################################################




# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2014-10-02 Added the dpwOtp class from lib/dpwotp.py
#             Cornelius Kölbel, <cornelius@privacyidea.org>
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  LSE
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
This file contains the definition of the tagespasswort token class
'''
from privacyidea.lib.util import getParam
from privacyidea.lib.log import log_with
from privacyidea.lib.tokenclass import TokenClass
from privacyidea.lib.config import getFromConfig
from privacyidea.lib.error import TokenAdminError
from hashlib import md5
from datetime import (datetime,
                      timedelta)
from binascii import hexlify

from privacyidea.lib.crypto import zerome

import logging
log = logging.getLogger(__name__)
optional = True
required = False


class dpwOtp:

    def __init__(self, secObj, digits=6):
        self.secretObject = secObj
        self.digits = digits

    def checkOtp(self, anOtpVal, window=0, options=None):
        '''
        window is the seconds before and after the current time
        '''
        res = -1

        key = self.secretObject.getKey()

        date_string = datetime.now().strftime("%d%m%y")
        input_data = key + date_string

        md = hexlify(md5(input_data).digest())
        md = md[len(md) - self.digits:]
        otp = int(md, 16)
        otp = unicode(otp)
        otp = otp[len(otp) - self.digits:]

        if unicode(anOtpVal) == otp:
            res = 1

        zerome(key)
        del key

        return res

    def getOtp(self, date_string=None):

        key = self.secretObject.getKey()

        if date_string is None:
            date_string = datetime.now().strftime("%d%m%y")

        input_data = key + date_string

        md = hexlify(md5(input_data).digest())
        md = md[len(md) - self.digits:]
        otp = int(md, 16)
        otp = unicode(otp)
        otp = otp[len(otp) - self.digits:]

        zerome(key)
        del key

        return otp


###############################################
class TagespasswortTokenClass(TokenClass):
    '''
    The Tagespasswort is a one time password that is calculated based on the
    day input.
    '''

    def __init__(self, aToken):
        TokenClass.__init__(self, aToken)
        self.setType(u"DPW")

        self.hKeyRequired = True

    @classmethod
    def getClassType(cls):
        return "dpw"

    @classmethod
    def getClassPrefix(cls):
        return "DOTP"

    @classmethod
    @log_with(log)
    def getClassInfo(cls, key=None, ret='all'):
        '''
        getClassInfo - returns a subtree of the token definition

        :param key: subsection identifier
        :type key: string

        :param ret: default return value, if nothing is found
        :type ret: user defined

        :return: subsection if key exists or user defined
        :rtype: s.o.

        '''
        res = {'type': 'dpw',
               'title': 'Tagespasswort Token',
               'description': ('A token uses a new password every day.'),
               'init': {'page': {'html': 'tagespassworttoken.mako',
                                 'scope': 'enroll', },
                        'title': {'html': 'tagespassworttoken.mako',
                                  'scope': 'enroll.title', },
                        },
               'config': {},
               'selfservice': {},
               'policy': {},
               }
        # I don't think we need to define the lost token policies here...

        if key is not None and key in res:
            ret = res.get(key)
        else:
            if ret == 'all':
                ret = res

        return ret

    def update(self, param):

        # check for the required parameters
        if (self.hKeyRequired is True):
            getParam(param, "otpkey", required)

        TokenClass.update(self, param)

    def reset(self):
        TokenClass.reset(self)

    @log_with(log)
    def checkOtp(self, anOtpVal, counter, window, options=None):
        res = -1

        try:
            otplen = int(self.token.privacyIDEAOtpLen)
        except ValueError:
            return res

        secretHOtp = self.token.getHOtpKey()

        dpw = dpwOtp(secretHOtp, otplen)
        res = dpw.checkOtp(anOtpVal, window=window)

        return res

    @log_with(log)
    def getOtp(self, curTime=None):
        # kay: init value
        res = (-1, 0, 0, 0)

        try:
            otplen = int(self.token.privacyIDEAOtpLen)
        except ValueError:
            return res

        secretHOtp = self.token.getHOtpKey()

        dpw = dpwOtp(secretHOtp, otplen)

        date_string = None
        if curTime:
            if type(curTime) == datetime:
                date_string = curTime.strftime("%d%m%y")
            elif type(curTime) == unicode:
                date_string = datetime.strptime(curTime,
                                                "%Y-%m-%d %H:%M:%S.%f"
                                                ).strftime("%d%m%y")
            else:
                log.error("invalid curTime: %r. "
                          "You need to specify a datetime" % type(curTime))
        otpval = dpw.getOtp(date_string)
        pin = self.token.getPin()
        combined = "%s%s" % (otpval, pin)
        if getFromConfig("PrependPin") == "True":
            combined = "%s%s" % (pin, otpval)

        return (1, pin, otpval, combined)

    @log_with(log)
    def get_multi_otp(self, count=0, epoch_start=0, epoch_end=0,
                      curTime=None, timestamp=None):
        '''
        This returns a dictionary of multiple future OTP values of the
        Tagespasswort token

        parameter
            count    - how many otp values should be returned
            epoch_start    - time based tokens: start when
            epoch_end      - time based tokens: stop when

        return
            True/False
            error text
            OTP dictionary
        '''
        otp_dict = {"type": "DPW", "otp": {}}
        ret = False
        error = "No count specified"
        try:
            otplen = int(self.token.privacyIDEAOtpLen)
        except ValueError as ex:
            log.error("%r" % ex)
            return (False, unicode(ex), otp_dict)

        secretHOtp = self.token.getHOtpKey()
        dpw = dpwOtp(secretHOtp, otplen)
        log.debug("retrieving %i OTP values for token %s" % (count, dpw))

        if count > 0:
            now = datetime.now()
            if curTime:
                if type(curTime) == datetime:
                    now = curTime
                elif type(curTime) == unicode:
                    now = datetime.strptime(curTime, "%Y-%m-%d %H:%M:%S.%f")
                else:
                    log.error("wrong curTime type: %s" % type(curTime))
                    raise TokenAdminError("[get_multi_otp] wrong curTime type:"
                                          " %s (%s)" %
                                          (type(curTime), curTime), id=2001)
            for i in range(count):
                delta = timedelta(days=i)
                date_string = (now + delta).strftime("%d%m%y")
                otpval = dpw.getOtp(date_string=date_string)
                otp_dict["otp"][(now + delta).strftime("%y-%m-%d")] = otpval
            ret = True

        return (ret, error, otp_dict)



# -*- coding: utf-8 -*-
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2014-10-02 Code cleanup for code coverage
#             Cornelius Kölbel, cornelius@privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
'''
This file is part of the privacyidea service
It can retrieve information of the token,
exspecially the OTP value.
Use with care. This controller is usually disabled.
'''


import logging

from pylons import tmpl_context as c
from pylons import request, response
from pylons import config
from pylons.templating import render_mako as render


from privacyidea.lib.base import BaseController

from privacyidea.lib.util import getParam
from privacyidea.lib.util import get_client
from privacyidea.lib.user import getUserFromParam
from privacyidea.lib.user import getDefaultRealm
from privacyidea.lib.user import getUserFromRequest

from privacyidea.lib.token import get_token_type_list
from privacyidea.lib.token import (getTokenType,
                                   getOtp,
                                   get_multi_otp,
                                   getTokens4UserOrSerial,
                                   getTokenRealms)
from privacyidea.lib.policy import PolicyClass, PolicyException
from privacyidea.lib.reply import sendResult, sendError
from privacyidea.lib.config import get_privacyIDEA_config
from privacyidea.model.meta import Session
from privacyidea.lib.log import log_with

import traceback

ENCODING = "utf-8"

optional = True
required = False

log = logging.getLogger(__name__)


class GettokenController(BaseController):

    '''
    The privacyidea.controllers are the implementation of the web-API to talk
    to the privacyIDEA server.
    The ValidateController is used to validate the username with its given
    OTP value.

    The Tagespasswort Token uses this controller to retrieve the current OTP
    value of     the Token and be able to set it in the application
    The functions of the GettokenController are invoked like this

        https://server/gettoken/<functionname>

    The functions are described below in more detail.
    '''
    @log_with(log)
    def __before__(self, action, **params):
        try:
            c.audit['client'] = get_client()
            tokentype = None
            if request.params.get('serial'):
                tokentype = getTokenType(request.params.get('serial'))
                
            self.Policy = PolicyClass(request, config, c,
                                      get_privacyIDEA_config(),
                                      tokentype=tokentype,
                                      token_type_list=get_token_type_list())
            self.before_identity_check(action)

        except Exception as exx:  # pragma: no cover
            log.error("%r exception %r" % (action, exx))
            log.error(traceback.format_exc())
            Session.rollback()
            Session.close()
            return sendError(response, exx, context='before')

        finally:
            pass

    @log_with(log)
    def __after__(self, action, **params):
        c.audit['administrator'] = getUserFromRequest(request).get("login")
        if 'serial' in request.params:
                c.audit['serial'] = request.params['serial']
                c.audit['token_type'] = getTokenType(request.params['serial'])
        self.audit.log(c.audit)
        
    @log_with(log)
    def getmultiotp(self, action, **params):
        '''
        This function is used to retrieve multiple otp values for a given user
        or a given serial. If the user has more than one token, the list of
        the tokens is returned.

        method:
            gettoken/getmultiotp

        
        :param serial: the serial number of the token
        :param count: number of otp values to return
        :param curTime: ONLY for internal testing: datetime.datetime object
        :type curTime: datetime object
        :param timestamp: the unix time
        :type timestamp: int
        
        :return: JSON response
        '''

        getotp_active = config.get("privacyideaGetotp.active")
        if getotp_active != "True":  # pragma: no cover
            return sendError(response, "getotp is not activated.", 0)

        param = request.params
        ret = {}

        try:
            serial = getParam(param, "serial", required)
            tokenrealms = getTokenRealms(serial)
            count = int(getParam(param, "count", required))
            curTime = getParam(param, "curTime", optional)
            timestamp = getParam(param, "timestamp", optional)
            view = getParam(param, "view", optional)

            r1 = self.Policy.checkPolicyPre('admin', 'getotp', param,
                                            tokenrealms=tokenrealms)
            log.debug("admin-getotp returned %s" % r1)

            max_count = self.Policy.checkPolicyPre('gettoken', 'max_count',
                                                   param,
                                                   tokenrealms=tokenrealms)
            log.debug("checkpolicypre returned %s" % max_count)
            if count > max_count:
                count = max_count

            log.debug("retrieving OTP value for token %s" % serial)
            ret = get_multi_otp(serial, count=int(count),
                                curTime=curTime,
                                timestamp=timestamp)
            ret["serial"] = serial

            c.audit['success'] = True
            Session.commit()

            if view:
                c.ret = ret
                return render('/manage/multiotp_view.mako')
            else:
                return sendResult(response, ret, 0)

        except PolicyException as pe:  # pragma: no cover
            log.error("gettoken/getotp policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe), 1)

        except Exception as e:  # pragma: no cover
            log.error("gettoken/getmultiotp failed: %r" % e)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, "gettoken/getmultiotp failed: %s"
                             % unicode(e), 0)

        finally:
            Session.close()

    @log_with(log)
    def getotp(self, action, **params):
        '''
        This function is used to retrieve the current otp value for a given
        user or a given serial
        If the user has more than one token, the list of the tokens is
        returned.

        method:
            gettoken/getotp

        arguments:
            user    - username / loginname
            realm   - additional realm to match the user to a useridresolver
            serial  - the serial number of the token
            curTime - used ONY for internal testing: datetime.datetime object

        returns:
            JSON response
        '''

        getotp_active = config.get("privacyideaGetotp.active")
        if getotp_active != "True":  # pragma: no cover
            return sendError(response, "getotp is not activated.", 0)

        param = request.params
        ret = {}
        res = -1
        otpval = ""
        passw = ""
        serials = []

        try:

            serial = getParam(param, "serial", optional)
            user = getUserFromParam(param, optional)
            curTime = getParam(param, "curTime", optional)

            c.audit['user'] = user.login
            
            if serial:
                log.debug("retrieving OTP value for token %s" % serial)
            elif user.login:
                c.audit['realm'] = user.realm or getDefaultRealm()
                log.debug("retrieving OTP value for token for user %s@%s" %
                          (user.login, user.realm))
                toks = getTokens4UserOrSerial(user, serial)
                tokennum = len(toks)
                if tokennum > 1:
                    log.debug("The user has more than one token. "
                              "Returning the list of serials")
                    res = -3
                    for token in toks:
                        serials.append(token.getSerial())
                elif 1 == tokennum:
                    serial = toks[0].getSerial()
                    # The policy needs the token type
                    self.Policy.set_tokentype(toks[0].getTokentype())
                    log.debug("retrieving OTP for token %s for user %s@%s" %
                              (serial, user.login, user.realm))
                else:
                    log.debug("no token found for user %s@%s" %
                              (user.login, user.realm))
                    res = -4
            else:
                res = -5

            # if a serial was given or a unique serial could be received
            # from the given user.
            if serial:
                tokenrealms = getTokenRealms(serial)
                max_count = self.Policy.checkPolicyPre('gettoken',
                                                       'max_count',
                                                       param,
                                                       tokenrealms=tokenrealms)
                log.debug("checkpolicypre returned %s" % max_count)
                if max_count <= 0:
                    return sendError(response,
                                     "The policy forbids receiving OTP values "
                                     "for the token %s in this realm" %
                                     serial,
                                     1)

                (res, pin, otpval, passw) = getOtp(serial, curTime=curTime)

            c.audit['success'] = True

            if int(res) < 0:
                ret['result'] = False
                if res == -3:
                    ret['description'] = "The user has more than one token"
                    ret['serials'] = serials
                if res == -4:
                    ret['description'] = "No Token found for this user"
                if res == -5:
                    ret['description'] = "you need to provide a user or a "\
                                         "serial"
            else:
                ret['result'] = True
                ret['otpval'] = otpval
                ret['pin'] = pin
                ret['pass'] = passw

            Session.commit()
            return sendResult(response, ret, 0)

        except PolicyException as pe:
            log.error("gettoken/getotp policy failed: %r" % pe)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response, unicode(pe), 1)

        except Exception as e:  # pragma: no cover
            log.error("gettoken/getotp failed: %r" % e)
            log.error(traceback.format_exc())
            Session.rollback()
            return sendError(response,
                             "gettoken/getotp failed: %s" % unicode(e), 0)

        finally:
            Session.close()


# eof######################################################################



# -*- coding: utf-8 -*-
import codecs

from pyparsing import Literal, White, Word, alphanums, CharsNotIn
from pyparsing import Forward, Group, Optional, OneOrMore, ZeroOrMore
from pyparsing import pythonStyleComment, Regex


class BaseParser(object):
    
    def get(self):
        """
        return the grouped config
        """
        return
    
    def get_dict(self):
        '''
        return the client config as a dictionary.
        '''
        ret = {}
        config = self.get()
        for client in config:
            client_config = {}
            for attribute in client[1]:
                client_config[attribute[0]] = attribute[1]
            ret[client[0]] = client_config
        return ret
    
    def dump(self):
        conf = self.get()
        for client in conf:
            print "%s: %s" % (client[0], client[1])

    def format(self, dict_config):
        '''
        :return: The formatted data as it would be written to a file
        '''
        return
    
    def save(self, dict_config=None, outfile=None):
        if not outfile:
            outfile = self.file
        if dict_config:
            output = self.format(dict_config)
            f = codecs.open(outfile, 'w', 'utf-8')
            for line in output.splitlines():
                f.write(line + "\n")
            f.close()


class ClientConfParser(BaseParser):

    key = Word(alphanums + "_")
    client_key = Word(alphanums + "-_/.:")
    space = White().suppress()
    value = CharsNotIn("{}\n# ")
    comment = ("#")
    assignment = (key
                  + Optional(space)
                  + Literal("=").suppress()
                  + Optional(space)
                  + value
                  + Optional(space)
                  + Optional("#"))
    client_block = Forward()
    client_block << Group((Literal("client").suppress()
                          + space
                          + client_key)
                          + Literal("{").suppress()
                          + Group(ZeroOrMore(Group(assignment)))
                          + Literal("}").suppress()
                          )
    
    client_file = OneOrMore(client_block).ignore(pythonStyleComment)
    
    file_header = """# File parsed and saved by privacyidea.\n\n"""
    
    def __init__(self,
                 infile="/etc/freeradius/clients.conf",
                 content=None):
        self.file = None
        if content:
            self.content = content
        else:
            self.file = infile
            self._read()

    def _read(self):
        """
        Reread the contents from the disk
        """
        f = codecs.open(self.file, "r", "utf-8")
        self.content = f.read()
        f.close()
        
    def get(self):
        """
        return the grouped config
        """
        if self.file:
            self._read()
        config = self.client_file.parseString(self.content)
        return config
    
    def format(self, dict_config):
        '''
        :return: The formatted data as it would be written to a file
        '''
        output = ""
        output += self.file_header
        for client, attributes in dict_config.iteritems():
            output += "client %s {\n" % client
            for k, v in attributes.iteritems():
                output += "    %s = %s\n" % (k, v)
            output += "}\n\n"
        return output


class UserConfParser(BaseParser):
    
    key = Word(alphanums + "-")
    username = Word(alphanums + "@_.-/")
    client_key = Word(alphanums + "-_/.:")
    space = White().suppress()
    comma = (",")
    value = CharsNotIn("{}\n#, ")
    comment = ("#")
    # operator = ":="
    operator = Regex(":=|==|=|\+=|!=|>|>=|<|<=|=~|!~|=\*|!\*")
    assignment = Group(space
                       + key
                       + space.suppress()
                       + operator
                       + space.suppress()
                       + value
                       + ZeroOrMore(space).suppress()
                       + ZeroOrMore(comma).suppress())
    user_block = Forward()
    # USERNAME key operator value
    # <tab> key operator value
    user_block << Group(username
                        + space
                        + key
                        + operator
                        + space
                        + value
                        + Group(ZeroOrMore(assignment)))
    
    user_file = OneOrMore(user_block).ignore(pythonStyleComment)
    
    file_header = """# File parsed and saved by privacyidea.\n\n"""
    
    def __init__(self,
                 infile="/etc/freeradius/users",
                 content=None):
        self.file = None
        if content:
            self.content = content
        else:
            self.file = infile
            f = codecs.open(self.file, "r", "utf-8")
            self.content = f.read()
            f.close()
            
    def get(self):
        """
        return the grouped config
        
        something like this:
        [
        ['DEFAULT', 'Framed-Protocol', '==', 'PPP', [['Framed-Protocol', '=', 'PPP'], ['Framed-Compression', '=', 'Van-Jacobson-TCP-IP']]],
        ['DEFAULT', 'Hint', '==', '"CSLIP"', [['Framed-Protocol', '=', 'SLIP'], ['Framed-Compression', '=', 'Van-Jacobson-TCP-IP']]],
        ['DEFAULT', 'Hint', '==', '"SLIP"', [['Framed-Protocol', '=', 'SLIP']]]
        ]
        """
        config = self.user_file.parseString(self.content)
        return config
    
    def format(self, config):
        '''
        :return: The formatted data as it would be written to a file
        '''
        output = ""
        output += self.file_header
        for user in config:
            output += "%s %s %s %s\n" % (user[0], user[1], user[2], user[3])
            if len(user[4]):
                i = 0
                for reply_item in user[4]:
                    i += 1
                    output += "\t%s %s %s" % (reply_item[0],
                                              reply_item[1],
                                              reply_item[2])
                    if i < len(user[4]):
                        output += ","
                    output += "\n"
            output += "\n"
        return output
        

def main():  # pragma: no cover
    CP = ClientConfParser()
    config = CP.get_dict()
    # Here we could mangle with the config...
    CP.save(config, infile="/etc/freeradius/clients.conf.new")
    

if __name__ == '__main__':
    main()